import gc
import json
import logging
import numpy as np
import os
import pandas as pd
from typing import Any, Dict, Optional, Text, List

from schematic.schemas import SchemaGenerator
from schematic.utils.io_utils import load_json

logger = logging.getLogger(__name__)

class AttributesExplorer():
    def __init__(self,
                 path_to_jsonld: str,
                 )-> None:
        
        self.path_to_jsonld = path_to_jsonld
        self.json_data_model = load_json(self.path_to_jsonld)
        self.jsonld = load_json(self.path_to_jsonld)

        # instantiate a schema generator to retrieve db schema graph from metadata model graph
        self.sg = SchemaGenerator(self.path_to_jsonld)

        self.output_path = self.create_output_path('merged_csv')
    
    def create_output_path(self, terminal_folder):
        ''' Create output path to store Observable visualization data if it does not already exist.
        
        Args: self.path_to_jsonld
        
        Returns: output_path (str): path to store outputs 
        '''
        base_dir = os.path.dirname(self.path_to_jsonld)
        self.schema_name = self.path_to_jsonld.split('/')[-1].split('.model.jsonld')[0]
        output_path = os.path.join(base_dir, 'visualization', self.schema_name, terminal_folder)
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        return output_path

    def convert_string_cols_to_json(self, df: pd.DataFrame, cols_to_modify: list):
        """Converts values in a column from strings to JSON list 
        for upload to Synapse.
        """
        for col in df.columns:
            if col in cols_to_modify:
                df[col] = df[col].apply(lambda x: json.dumps([y.strip() for y in x]) if x != "NaN" and x  and x == np.nan else x)
        return df

    def parse_attributes(self, save_file=True):
        '''
        Args: save_file (bool):
                True: merged_df is saved locally to output_path.
                False: merged_df is returned.

        Returns:
            merged_df (pd.DataFrame): dataframe containing data relating to attributes
                for the provided data model for all components in the data model. 
                Dataframe is saved locally as a csv if save_file == True, or returned if
                save_file == False. 
        Raises:
            ValueError:
                If unable hits an error while attempting to get conditional requirements. 
                This error is likely to be found if there is a mismatch in naming.
        '''
        # get all components
        component_dg = self.sg.se.get_digraph_by_edge_type('requiresComponent')
        components = component_dg.nodes()
        # For each data type to be loaded gather all attribtes the user would
        # have to provide.

        df_store = []
        for component in components:
            data_dict = {}
            # get the json schema
            json_schema = self.sg.get_json_schema_requirements(
                source_node=component, schema_name=self.path_to_jsonld)

            # Gather all attribues, their valid values and requirements
            for key, value in json_schema['properties'].items():
                data_dict[key] = {}
                for k, v in value.items():
                    if k == 'enum':
                        data_dict[key]['Valid Values'] = value['enum']
                if key in json_schema['required']:
                    data_dict[key]['Required'] = True
                else:
                    data_dict[key]['Required'] = False
                data_dict[key]['Component'] = component
            # Add additional details per key (from the JSON-ld)
            for dic in self.jsonld['@graph']:
                if 'sms:displayName' in dic.keys():
                    key = dic['sms:displayName']
                    if key in data_dict.keys():
                        data_dict[key]['Attribute'] = dic['sms:displayName']
                        data_dict[key]['Label'] = dic['rdfs:label']
                        data_dict[key]['Description'] = dic['rdfs:comment']
                        if 'validationRules' in dic.keys():
                            data_dict[key]['Validation Rules'] = dic['validationRules']
            # Find conditional dependencies
            if 'allOf' in json_schema.keys():
                for conditional_dependencies in json_schema['allOf']:
                    key = list(conditional_dependencies['then']['properties'])[0]
                    try:
                        if key in data_dict.keys():
                            if 'Cond_Req' not in data_dict[key].keys():
                                data_dict[key]['Cond_Req'] = []
                                data_dict[key]['Conditional Requirements'] = []
                            attribute = list(conditional_dependencies['if']['properties'])[0]
                            value = conditional_dependencies['if']['properties'][attribute]['enum']
                            # Capitalize attribute if it begins with a lowercase letter, for aesthetics.
                            if attribute[0].islower():
                                attribute = attribute.capitalize()

                            # Remove "Type" (i.e. turn "Biospecimen Type" to "Biospcimen")
                            if "Type" in attribute: 
                                attribute = attribute.split(" ")[0]
                            
                            # Remove "Type" (i.e. turn "Tissue Type" to "Tissue")
                            if "Type" in value[0]:
                                value[0] = value[0].split(" ")[0]

                            conditional_statement = f'{attribute} is "{value[0]}"'
                            if conditional_statement not in data_dict[key]['Conditional Requirements']:
                                data_dict[key]['Cond_Req'] = True
                                data_dict[key]['Conditional Requirements'].extend([conditional_statement])
                    except:
                        raise ValueError(
                            f"There is an error getting conditional requirements related "
                            "to the attribute: {key}. The error is likely caused by naming inconsistencies (e.g. uppercase, camelcase, ...)"
                        )

            for key, value in data_dict.items():
                if 'Conditional Requirements' in value.keys():

                    ## reformat conditional requirement 

                    # get all attributes 
                    attr_lst = [i.split(" is ")[-1] for i in data_dict[key]['Conditional Requirements']]
                    
                    # join a list of attributes by using OR 
                    attr_str = " OR ".join(attr_lst)

                    # reformat the conditional requirement 
                    component_name = data_dict[key]['Conditional Requirements'][0].split(' is ')[0]
                    conditional_statement_str = f' If {component_name} is {attr_str} then "{key}" is required'

                    data_dict[key]['Conditional Requirements'] = conditional_statement_str
            df = pd.DataFrame(data_dict)
            df = df.T
            cols = ['Attribute', 'Label', 'Description', 'Required', 'Cond_Req', 'Valid Values', 'Conditional Requirements', 'Validation Rules', 'Component']
            cols = [col for col in cols if col in df.columns]
            df = df[cols]
            df = self.convert_string_cols_to_json(df, ['Valid Values'])
            #df.to_csv(os.path.join(csv_output_path, data_type + '.vis_data.csv'))
            df_store.append(df)

        merged_attributes_df = pd.concat(df_store, join='outer')
        cols = ['Attribute', 'Label', 'Description', 'Required', 'Cond_Req', 'Valid Values', 'Conditional Requirements', 'Validation Rules', 'Component']
        cols = [col for col in cols if col in merged_attributes_df.columns]

        merged_attributes_df = merged_attributes_df[cols]
        if save_file == True:
            return merged_attributes_df.to_csv(os.path.join(self.output_path, self.schema_name + 'attributes_data.vis_data.csv'))
        elif save_file == False:
            return merged_attributes_df.to_csv()



from errno import ENOENT
from os import pathsep
from re import split
from pkg_resources import (
    resource_exists,
    resource_filename,
    resource_stream,
    resource_string,
    resource_listdir,
)


class InvalidResourceError(Exception):
    """
    Args:
        uri {String}: The URI which was requested within the given loader's
        that did not exist or was malformed.
    """

    def __init__(self, namespace, requested_uri):
        self.namespace = namespace
        self.requested_uri = requested_uri
        self.message = "Resource does not exist or is declared incorrectly"
        self.errno = ENOENT
        super(InvalidResourceError, self).__init__(self.message)

    def __str__(self):
        return '{}({}), "{}" of {}'.format(
            self.message, self.errno, self.requested_uri, self.namespace
        )

    def __repr__(self):
        return self.__str__()


class Loader(object):
    """
    Args:
    namespace {String}: The namespace within the package (relative to the package root)
    to load resources from. Using the magic variable __name__ is suggested as when the script
    is run as "__main__" it will load the most recent local resources instead of the cached
    egg resources.

    prefix {String}: Set a prefix for all URIs. Use a prefix if resources are centrally
    located in a single place the uri's will be prefixed automatically by the loader.
    """

    def __init__(self, namespace, **opts):
        self.namespace = namespace
        self.prefix = opts.get("prefix", "")
        self.local = opts.get("local", False)

        if not self.local:
            self.namespace = split(r"\.|\\|\/", self.namespace)[0]

    def _resolve(self, uri):
        resource_uri = "/".join([self.prefix] + uri.split(pathsep))
        ns = self.namespace

        if not resource_exists(ns, resource_uri):
            raise InvalidResourceError(ns, resource_uri)

        return ns, resource_uri

    def read(self, uri):
        """
        Read entire contents of resource. Same as open('path...').read()

        Args:
            uri {String}: URI of the resource.
        """
        ns, uri = self._resolve(uri)
        return resource_string(ns, uri)

    def open(self, uri):
        """
        Open a file object like handle to the resource. Same as open('path...')

        Args:
            uri {String}: URI of the resource.
        """
        ns, uri = self._resolve(uri)
        return resource_stream(ns, uri)

    def filename(self, uri):
        """
        Return the "most correct" filename for a resource. Same as os.path.normpath('path...')

        Args:
            uri {String}: URI of the resource.
        """
        ns, uri = self._resolve(uri)
        return resource_filename(ns, uri)

    def list(self, url):
        """
        Return a list of all resources within the given URL

        Args:
            url {String}: URL of the resources.
        """
        ns, uri = self._resolve(url)
        return map(lambda x: url + "/" + x, resource_listdir(ns, uri))


# call Loader() and pass `schematic`, which is the global package namespace
LOADER = Loader("schematic", prefix="etc")



import json
import logging
import string

import numpy as np
import pandas as pd
import re
import networkx as nx
from jsonschema import Draft7Validator, exceptions, validate, ValidationError, FormatError
from os.path import exists

# allows specifying explicit variable types
from typing import Any, Dict, Optional, Text, List

# handle schema logic; to be refactored as SchemaExplorer matures into a package
# as collaboration with Biothings progresses

from schematic.schemas.explorer import SchemaExplorer
from schematic.manifest.generator import ManifestGenerator
from schematic.schemas.generator import SchemaGenerator

#TODO: This module should only be aware of the store interface
# we shouldn't need to expose Synapse functionality explicitly
from schematic.store.synapse import SynapseStorage

from schematic.utils.df_utils import load_df

from schematic.models.validate_attribute import ValidateAttribute
from schematic.models.validate_manifest import validate_all


logger = logging.getLogger(__name__)


class MetadataModel(object):
    """Metadata model wrapper around schema.org specification graph.

    Provides basic utilities to:

    1) manipulate the metadata model
    2) generate metadata model views:
        - generate manifest view of the metadata model
        - generate validation schema view of the metadata model
    """

    def __init__(self, inputMModelLocation: str, inputMModelLocationType: str,) -> None:

        """Instantiates a MetadataModel object.

        Args:
            inputMModelLocation: local path, uri, synapse entity id (e.g. gs://, syn123, /User/x/…); present location
            inputMModelLocationType: specifier to indicate where the metadata model resource can be found (e.g. 'local' if file/JSON-LD is on local machine)
        """
        # extract extension of 'inputMModelLocation'
        # ensure that it is necessarily pointing to a '.jsonld' file
        if inputMModelLocation.rpartition(".")[-1] == "jsonld":
            logger.debug(
                f"Initializing SchemaGenerator object from {inputMModelLocation} schema."
            )
            self.inputMModelLocation = inputMModelLocation

            self.sg = SchemaGenerator(inputMModelLocation)
        else:
            raise TypeError(
                f"Please make sure {inputMModelLocation} is a .jsonld file."
            )

        # check if the type of MModel file is "local"
        # currently, the application only supports reading from local JSON-LD files
        if inputMModelLocationType == "local":
            self.inputMModelLocationType = inputMModelLocationType
        else:
            raise ValueError(
                f"The type '{inputMModelLocationType}' is currently not supported."
            )

    def getModelSubgraph(self, rootNode: str, subgraphType: str) -> nx.DiGraph:
        """Gets a schema subgraph from rootNode descendants based on edge/node properties of type subgraphType.

        Args:
            rootNode: a schema node label (i.e. term).
            subgraphType: the kind of subgraph to traverse (i.e. based on node properties or edge labels).

        Returns:
            A directed subgraph (networkx DiGraph) of the metadata model with vertex set root node descendants.

        Raises:
            ValueError: rootNode not found in metadata model.
        """
        pass

    def getOrderedModelNodes(self, rootNode: str, relationshipType: str) -> List[str]:
        """Get a list of model objects ordered by their topological sort rank in a model subgraph on edges of a given relationship type.

        Args:
            rootNode: a schema object/node label (i.e. term)
            relationshipType: edge label type of the schema subgraph (e.g. requiresDependency)

        Returns:
            An ordered list of objects, that are all descendants of rootNode.

        Raises:
            ValueError: rootNode not found in metadata model.
        """
        ordered_nodes = self.sg.get_descendants_by_edge_type(
            rootNode, relationshipType, connected=True, ordered=True
        )

        ordered_nodes.reverse()

        return ordered_nodes

    def getModelManifest(
        self,
        title: str,
        rootNode: str,
        datasetId: str = None,
        jsonSchema: str = None,
        filenames: list = None,
        useAnnotations: bool = False,
        sheetUrl: bool = True,
    ) -> str:
        """Gets data from the annotations manifest file.

        TBD: Does this method belong here or in manifest generator?

        Args:
            rootNode: a schema node label (i.e. term).
            useAnnotations: whether to populate manifest with current file annotations (True) or not (False, default).

        Returns:
            A manifest URI (assume Google doc for now).

        Raises:
            ValueError: rootNode not found in metadata model.
        """
        additionalMetadata = {}
        if filenames:
            additionalMetadata["Filename"] = filenames

        mg = ManifestGenerator(
            path_to_json_ld=self.inputMModelLocation,
            title=title,
            root=rootNode,
            additional_metadata=additionalMetadata,
            use_annotations=useAnnotations,
        )

        if datasetId:
            return mg.get_manifest(
                dataset_id=datasetId, json_schema=jsonSchema, sheet_url=sheetUrl
            )

        return mg.get_manifest(sheet_url=sheetUrl)

    def get_component_requirements(
        self, source_component: str, as_graph: bool = False
    ) -> List:
        """Given a source model component (see https://w3id.org/biolink/vocab/category for definnition of component), return all components required by it.
        Useful to construct requirement dependencies not only between specific attributes but also between categories/components of attributes;
        Can be utilized to track metadata completion progress across multiple categories of attributes.

        Args:
            source_component: an attribute label indicating the source component.
            as_graph: if False return component requirements as a list; if True return component requirements as a dependency graph (i.e. a DAG)

        Returns:
            A list of required components associated with the source component.
        """

        # get required components for the input/source component
        req_components = self.sg.get_component_requirements(source_component)

        # retreive components as graph
        if as_graph:
            req_components_graph = self.sg.get_component_requirements_graph(
                source_component
            )

            # serialize component dependencies DAG to a edge list of node tuples
            req_components = list(req_components_graph.edges())

            return req_components

        return req_components

    # TODO: abstract validation in its own module
    def validateModelManifest(
        self, manifestPath: str, rootNode: str, restrict_rules: bool = False, jsonSchema: str = None, project_scope: List = None,
    ) -> List[str]:
        """Check if provided annotations manifest dataframe satisfies all model requirements.

        Args:
            rootNode: a schema node label (i.e. term).
            manifestPath: a path to the manifest csv file containing annotations.
            restrict_rules: bypass great expectations and restrict rule options to those implemented in house

        Returns:
            A validation status message; if there is an error the message.
            contains the manifest annotation record (i.e. row) that is invalid, along with the validation error associated with this record.

        Raises:
            ValueError: rootNode not found in metadata model.
        """
        # get validation schema for a given node in the data model, if the user has not provided input validation schema
        
        if not jsonSchema:
            jsonSchema = self.sg.get_json_schema_requirements(
                rootNode, rootNode + "_validation"
            )

        errors = []
        warnings = []

        load_args={
            "dtype":"string",
            }
        # get annotations from manifest (array of json annotations corresponding to manifest rows)
        manifest = load_df(
            manifestPath, preserve_raw_input=False, **load_args,
        )  # read manifest csv file as is from manifest path

        # handler for mismatched components/data types
        # throw TypeError if the value(s) in the "Component" column differ from the selected template type
        if ("Component" in manifest.columns) and (
            (len(manifest["Component"].unique()) > 1)
            or (manifest["Component"].unique()[0] != rootNode)
        ):
            logging.error(
                f"The 'Component' column value(s) {manifest['Component'].unique()} do not match the "
                f"selected template type '{rootNode}'."
            )

            # row indexes for all rows where 'Component' is rootNode
            row_idxs = manifest.index[manifest["Component"] != rootNode].tolist()
            # column index value for the 'Component' column
            col_idx = manifest.columns.get_loc("Component")
            # Series with index and 'Component' values from manifest
            mismatched_ser = manifest.iloc[row_idxs, col_idx]
            for index, component in mismatched_ser.items():
                errors.append(
                    [
                        index + 2,
                        "Component",
                        f"Component value provided is: '{component}', whereas the Template Type is: '{rootNode}'",
                        # tuple of the component in the manifest and selected template type
                        # check: R/Reticulate cannnot handle dicts? So returning tuple
                        (component, rootNode),
                    ]
                )

            return errors, warnings

        errors, warnings, manifest = validate_all(self, errors, warnings, manifest, manifestPath, self.sg, jsonSchema, restrict_rules, project_scope)
        return errors, warnings

    def populateModelManifest(self, title, manifestPath: str, rootNode: str, return_excel = False) -> str:
        """Populate an existing annotations manifest based on a dataframe.
            TODO: Remove this method; always use getModelManifest instead

        Args:
            rootNode: a schema node label (i.e. term).
            manifestPath: a path to the manifest csv file containing annotations.

        Returns:
            A link to the filled in model manifest (e.g. google sheet).

        Raises:
            ValueError: rootNode not found in metadata model.
        """
        mg = ManifestGenerator(
            path_to_json_ld=self.inputMModelLocation, title=title, root=rootNode
        )

        emptyManifestURL = mg.get_manifest()

        return mg.populate_manifest_spreadsheet(manifestPath, emptyManifestURL, return_excel = return_excel, title=title)

    def submit_metadata_manifest(
        self,
        manifest_path: str,
        path_to_json_ld: str,
        dataset_id: str,
        manifest_record_type: str,
        restrict_rules: bool,
        validate_component: str = None,
        use_schema_label: bool = True,
        hide_blanks: bool = False,
        input_token: str = None,
        project_scope: List = None,
    ) -> string:
        """Wrap methods that are responsible for validation of manifests for a given component, and association of the
        same manifest file with a specified dataset.
        Args:
            manifest_path: Path to the manifest file, which contains the metadata.
            dataset_id: Synapse ID of the dataset on Synapse containing the metadata manifest file.
            validate_component: Component from the schema.org schema based on which the manifest template has been generated.
        Returns:
            Manifest ID: If both validation and association were successful.
        Exceptions:
            ValueError: When validate_component is provided, but it cannot be found in the schema.
            ValidationError: If validation against data model was not successful.
        """

        #TODO: avoid explicitly exposing Synapse store functionality
        # just instantiate a Store class and let it decide at runtime/config
        # the store type
        syn_store = SynapseStorage(input_token = input_token, project_scope = project_scope)
        manifest_id=None
        censored_manifest_id=None
        restrict_maniest=False
        censored_manifest_path=manifest_path.replace('.csv','_censored.csv')
        # check if user wants to perform validation or not
        if validate_component is not None:

            try:
                # check if the component ("class" in schema) passed as argument is valid (present in schema) or not
                self.sg.se.is_class_in_schema(validate_component)
            except:
                # a KeyError exception is raised when validate_component fails in the try-block above
                # here, we are suppressing the KeyError exception and replacing it with a more
                # descriptive ValueError exception
                raise ValueError(
                    f"The component '{validate_component}' could not be found "
                    f"in the schema here '{path_to_json_ld}'"
                )

            # automatic JSON schema generation and validation with that JSON schema
            val_errors, val_warnings = self.validateModelManifest(
                manifestPath=manifest_path, rootNode=validate_component, restrict_rules=restrict_rules, project_scope=project_scope,
            )

            # if there are no errors in validation process
            if val_errors == []:                
                # upload manifest file from `manifest_path` path to entity with Syn ID `dataset_id`
                if exists(censored_manifest_path):
                    censored_manifest_id = syn_store.associateMetadataWithFiles(
                        schemaGenerator = self.sg,
                        metadataManifestPath = censored_manifest_path,
                        datasetId = dataset_id, 
                        manifest_record_type = manifest_record_type,
                        useSchemaLabel = use_schema_label,
                        hideBlanks = hide_blanks,
                    )
                    restrict_maniest = True
                
                manifest_id = syn_store.associateMetadataWithFiles(
                    schemaGenerator = self.sg,
                    metadataManifestPath = manifest_path, 
                    datasetId = dataset_id, 
                    manifest_record_type = manifest_record_type,
                    useSchemaLabel = use_schema_label, 
                    hideBlanks = hide_blanks,
                    restrict_manifest=restrict_maniest,
                )

                logger.info(f"No validation errors occured during validation.")
                return manifest_id
                
            else:
                raise ValidationError(
                    "Manifest could not be validated under provided data model. "
                    f"Validation failed with the following errors: {val_errors}"
                )

        # no need to perform validation, just submit/associate the metadata manifest file
        if exists(censored_manifest_path):
            censored_manifest_id = syn_store.associateMetadataWithFiles(
                schemaGenerator = self.sg,
                metadataManifestPath=censored_manifest_path,
                datasetId=dataset_id,
                manifest_record_type=manifest_record_type,
                useSchemaLabel=use_schema_label,
                hideBlanks=hide_blanks,
            )
            restrict_maniest = True
        
        manifest_id = syn_store.associateMetadataWithFiles(
            schemaGenerator = self.sg,
            metadataManifestPath=manifest_path,
            datasetId=dataset_id,
            manifest_record_type=manifest_record_type,
            useSchemaLabel=use_schema_label,
            hideBlanks=hide_blanks,
            restrict_manifest=restrict_maniest,
        )

        logger.debug(
            "Optional validation was not performed on manifest before association."
        )

        return manifest_id



class BaseStorage(object):
    """Common interface for data store back-ends (like Synapse and Gen3)."""

    def __init__(self):
        """Base function for initializing object."""
        raise NotImplementedError()



from io import StringIO
import json
import logging
import networkx as nx
import numpy as np
import os
from os import path
import pandas as pd

# allows specifying explicit variable types
from typing import Any, Dict, Optional, Text, List

from schematic.utils.viz_utils import visualize
from schematic.visualization.attributes_explorer import AttributesExplorer
from schematic.schemas.explorer import SchemaExplorer
from schematic.schemas.generator import SchemaGenerator
from schematic import LOADER
from schematic.utils.io_utils import load_json
from copy import deepcopy

# Make sure to have newest version of decorator


logger = logging.getLogger(__name__)
#OUTPUT_DATA_DIR = str(Path('tests/data/visualization/AMPAD').resolve())
#DATA_DIR = str(Path('tests/data').resolve())

class TangledTree(object):
    """
    """

    def __init__(self,
                 path_to_json_ld: str,
                 figure_type: str,
                 ) -> None:
        # Load jsonld
        self.path_to_json_ld = path_to_json_ld
        self.json_data_model = load_json(self.path_to_json_ld)

        # Parse schema name
        self.schema_name = path.basename(self.path_to_json_ld).split(".model.jsonld")[0]

        # Instantiate a schema generator to retrieve db schema graph from metadata model graph
        self.sg = SchemaGenerator(self.path_to_json_ld)      

        # Get metadata model schema graph
        self.G = self.sg.se.get_nx_schema()

        # Set Parameters
        self.figure_type = figure_type.lower()
        self.dependency_type = ''.join(('requires', self.figure_type.capitalize()))

        # Get names
        self.schema = load_json(self.path_to_json_ld)
        self.schema_abbr = self.schema_name.split('_')[0]

        # Initialize AttributesExplorer
        self.ae = AttributesExplorer(self.path_to_json_ld)

        # Create output paths.
        self.text_csv_output_path = self.ae.create_output_path('text_csv')
        self.json_output_path = self.ae.create_output_path('tangled_tree_json')

    def strip_double_quotes(self, string):
        # Remove double quotes from beginning and end of string.
        if string.startswith('"') and string.endswith('"'):
            string = string[1:-1]
        # now remove whitespace
        string = "".join(string.split())
        return string

    def get_text_for_tangled_tree(self, text_type, save_file=False):
        '''Gather the text that needs to be either higlighted or plain for the tangled tree visualization.
        Args:
            text_type (str): Choices = ['highlighted', 'plain'], determines the type of text
                rendering to return.
            save_file (bool): Determines if the outputs should be saved to disk or returned.
        Returns:
            If save_file==True: Saves plain or highlighted text as a CSV (to disk).
               save_file==False: Returns plain or highlighted text as a csv string.
        '''
        # Get nodes in the digraph, many more nodes returned if figure type is dependency
        cdg = self.sg.se.get_digraph_by_edge_type(self.dependency_type)
        nodes = cdg.nodes()

        if self.dependency_type == 'requiresComponent':
            component_nodes = nodes
        else:
            # get component nodes if making dependency figure
            component_dg = self.sg.se.get_digraph_by_edge_type('requiresComponent')
            component_nodes = component_dg.nodes()

        # Initialize lists
        highlighted = []
        plain = []

        # For each component node in the tangled tree gather the plain and higlighted text.
        for node in component_nodes:
            # Get the highlighted components based on figure_type
            if self.figure_type == 'component':
                highlight_descendants = self.sg.se.get_descendants_by_edge_type(node, 'requiresComponent')
            elif self.figure_type == 'dependency':
                highlight_descendants = [node]

            
            # Format text to be higlighted and gather text to be formated plain.
            if not highlight_descendants:
                # If there are no highlighted descendants just highlight the selected node (format for observable.)
                highlighted.append([node, "id", node])
                # Gather all the text as plain text.
                plain_descendants = [n for n in nodes if n != node]
            else:
                # Format higlighted text for Observable.
                for hd in highlight_descendants:
                    highlighted.append([node, "id", hd])
                # Gather the non-higlighted text as plain text descendants.
                plain_descendants = [node for node in nodes if node not in highlight_descendants]
            
            # Format all the plain text for observable.
            for nd in plain_descendants:
                plain.append([node, "id", nd])

        # Prepare df depending on what type of text we need.
        df = pd.DataFrame(locals()[text_type.lower()], columns = ['Component', 'type', 'name'])

        # Depending on input either export csv locally to disk or as a string.
        if save_file==True:
            file_name = f"{self.schema_abbr}_{self.figure_type}_{text_type}.csv"
            df.to_csv(os.path.join(self.text_csv_output_path, file_name))
            return
        elif save_file==False:
            return df.to_csv()

    def get_topological_generations(self):
        ''' Gather topological_gen, nodes and edges based on figure type.
        Outputs:
            topological_gen (List(list)):list of lists. Indicates layers of nodes.
            nodes: (Networkx NodeView) Nodes of the component or dependency graph. When iterated over it functions like a list.
            edges: (Networkx EdgeDataView) Edges of component or dependency graph. When iterated over it works like a list of tuples.
        '''
        # Get nodes in the digraph
        digraph = self.sg.se.get_digraph_by_edge_type(self.dependency_type)
        nodes = digraph.nodes()

        # Get subgraph
        mm_graph = self.sg.se.get_nx_schema()
        subg = self.sg.get_subgraph_by_edge_type(mm_graph, self.dependency_type)

        # Get edges and topological_gen based on figure type.
        if self.figure_type == 'component':
            edges = digraph.edges()
            topological_gen = list(reversed(list(nx.topological_generations(subg))))

        elif self.figure_type == 'dependency':
            rev_digraph = nx.DiGraph.reverse(digraph)
            edges = rev_digraph.edges()
            topological_gen = list(nx.topological_generations(subg))
        
        return topological_gen, nodes, edges, subg

    def remove_unwanted_characters_from_conditional_statement(self, cond_req: str) -> str:
        '''Remove unwanted characters from conditional statement
        Example of conditional requirement: If File Format IS "BAM" OR "CRAM" OR "CSV/TSV" then Genome Build is required
        Example output: File Format IS "BAM" OR "CRAM" OR "CSV/TSV"
        '''
        if "then" in cond_req:
            # remove everything after "then"
            cond_req_new = cond_req.split('then')[0]

            # remove "If" and empty space
            cond_req = cond_req_new.replace("If", "").lstrip().rstrip()
        return cond_req

    def get_ca_alias(self, conditional_requirements: list) -> dict:
        '''Get the alias for each conditional attribute. 

        NOTE: Obtaining attributes(attr) and aliases(ali) in this function is specific to how formatting
            is set in AttributesExplorer. If that formatting changes, this section
            will likely break or in the worst case have a silent error.         
        Input:
            conditional_requirements_list (list): list of strings of conditional requirements from outputs of AttributesExplorer.
        Output:
            ca_alias (dict):
                key: alias (attribute response)
                value: attribute
        '''
        ca_alias = {}

        # clean up conditional requirements
        conditional_requirements = [self.remove_unwanted_characters_from_conditional_statement(req) for req in conditional_requirements]

        for i, req in enumerate(conditional_requirements):
            if "OR" not in req:
                attr, ali = req.split(' is ')
                attr = "".join(attr.split())
                ali = self.strip_double_quotes(ali)
                ca_alias[ali] = attr
            else:
                attr, alias_str = req.split(' is ')
                alias_lst = alias_str.split(' OR ')
                for elem in alias_lst:
                    elem = self.strip_double_quotes(elem)
                    ca_alias[elem] = attr
        return ca_alias

    def gather_component_dependency_info(self, cn, attributes_df):
        '''Gather all component dependency information.
        Inputs:
            cn: (str) component name
            attributes_df: (Pandas DataFrame) Details for all attributes across all components. From AttributesExplorer.
        Outputs:
            conditional_attributes (list): List of conditional attributes for a particular component
            ca_alias (dict):
                key: alias (attribute response)
                value: attribute
            all_attributes (list): all attributes associated with a particular component.
        '''

        # Gather all component dependency information
        component_attributes = self.sg.get_descendants_by_edge_type(
                                        cn,
                                        self.dependency_type,
                                        connected=True
                                        )
        
        # Dont want to display `Component` in the figure so remove
        if 'Component' in component_attributes:
            component_attributes.remove('Component')
        
        # Gather conditional attributes so they can be added to the figure.
        if 'Cond_Req' in attributes_df.columns:
            conditional_attributes = list(attributes_df[(attributes_df['Cond_Req']==True)
                &(attributes_df['Component']==cn)]['Label'])
            ca_df = attributes_df[(attributes_df['Cond_Req']==True)&(attributes_df['Component']==cn)]
            conditional_requirements = list(attributes_df[(attributes_df['Cond_Req']==True)
                &(attributes_df['Component']==cn)]['Conditional Requirements'])
            ca_alias = self.get_ca_alias(conditional_requirements)
        else:
            # If there are no conditional attributes/requirements, initialize blank lists.
            conditional_attributes = []
            ca_alias = {}
        
        # Gather a list of all attributes for the current component.
        all_attributes = list(np.append(component_attributes,conditional_attributes))
        
        return conditional_attributes, ca_alias, all_attributes

    def find_source_nodes(self, nodes, edges, all_attributes=[]):
        '''Find all nodes in the graph that do not have a parent node.
        Inputs:
            nodes: (Networkx NodeView) Nodes of the component or dependency graph. When iterated over it functions like a list.
            edges: (Networkx EdgeDataView) Edges of component or dependency graph. When iterated over it works like a list of tuples.
            attributes_df: (Pandas DataFrame) Details for all attributes across all components. From AttributesExplorer.

        Outputs:
            source_nodes (list(str)): List of parentless nodes in 
        '''
        # Find edges that are not source nodes.
        not_source = []
        for node in nodes:
            for edge_pair in edges:
                if node == edge_pair[0]:
                    not_source.append(node)

        # Find source nodes as nodes that are not in not_source.
        source_nodes = []
        for node in nodes:
            if self.figure_type == 'dependency':
                if node not in not_source and node in all_attributes:
                    source_nodes.append(node)
            else:
                if node not in not_source:
                    source_nodes.append(node)
        return source_nodes

    def get_parent_child_dictionary(self, nodes, edges, all_attributes=[]):
        '''Based on the dependency type, create dictionaries between parent and child and child and parent attributes.
        Input:
            nodes: (Networkx NodeView) Nodes of the component or dependency graph.
            edges: (Networkx EdgeDataView (component figure) or List(list) (dependency figure))
                Edges of component or dependency graph.
            all_attributes:
        Output:
            child_parents (dict):
                key: child
                value: list of the childs parents
            parent_children (dict):
                key: parent
                value: list of the parents children
        '''
        child_parents = {}
        parent_children = {}

        if self.dependency_type == 'requiresComponent':
            
            # Construct child_parents dictionary
            for edge in edges:
                
                # Add child as a key
                if edge[0] not in child_parents.keys():
                    child_parents[edge[0]] = []
                
                # Add parents to list
                child_parents[edge[0]].append(edge[1])
            
            # Construct parent_children dictionary
            for edge in edges:
                
                # Add parent as a key
                if edge[1] not in parent_children.keys():
                    parent_children[edge[1]] = []
                
                # Add children to list
                parent_children[edge[1]].append(edge[0])
        
        elif self.dependency_type == 'requiresDependency':
            
            # Construct child_parents dictionary
            for edge in edges:
                
                # Check if child is an attribute for the current component
                if edge[0] in all_attributes:
                    
                    # Add child as a key
                    if edge[0] not in child_parents.keys():
                        child_parents[edge[0]] = []
                    
                    # Add parent to list if it is an attriute for the current component
                    if edge[1] in all_attributes:
                        child_parents[edge[0]].append(edge[1])
            
            # Construct parent_children dictionary
            for edge in edges:
                
                # Check if parent is an attribute for the current component
                if edge[1] in all_attributes:
                    
                    # Add parent as a key
                    if edge[1] not in parent_children.keys():
                        parent_children[edge[1]] = []
                    
                    # Add child to list if it is an attriute for the current component
                    if edge[0] in all_attributes:
                        parent_children[edge[1]].append(edge[0])

        return child_parents, parent_children

    def alias_edges(self, ca_alias:dict, edges) -> List[list]:
        '''Create new edges based on aliasing between an attribute and its response.
        Purpose:
            Create aliased edges.
            For example: 
                If BiospecimenType (attribute) is AnalyteBiospecimenType (response)
                Then ShippingConditionType (conditional requirement) is now required.
            In the model the edges that connect these options are:
                (AnalyteBiospecimenType, BiospecimenType)
                (ShippingConditionType, AnalyteBiospecimenType)
            Use alias defined in self.get_ca_alias along to define new edges that would 
            directly link attributes to their conditional requirements, in this 
            example the new edge would be:
                [ShippingConditionType, BiospecimenType]
        Inputs:
            ca_alias (dict):
                key: alias (attribute response)
                value: attribute
            edges (Networkx EdgeDataView): Edges of component or dependency graph. When iterated over it works like a list of tuples.
        Output:
            aliased_edges (List[lists]) of aliased edges.
        '''
        aliased_edges = []
        for i, edge in enumerate(edges):
            
            # construct one set of edges at a time
            edge_set = []
            
            # If the first edge has an alias add alias to the first position in the current edge set
            if edge[0] in ca_alias.keys():
                edge_set.append(ca_alias[edge[0]])
            
            # Else add the non-aliased edge
            else:
                edge_set.append(edge[0])

            # If the secod edge has an alias add alias to the first position in the current edge set
            if edge[1] in ca_alias.keys():
                edge_set.append(ca_alias[edge[1]])
            
            # Else add the non-aliased edge
            else:
                edge_set.append(edge[1])

            # Add new edge set to a the list of aliased edges.
            aliased_edges.append(edge_set)

        return aliased_edges

    def prune_expand_topological_gen(self, topological_gen, all_attributes, conditional_attributes):
        '''
        Purpose:
            Remake topological_gen with only relevant nodes.
                This is necessary since for the figure this function is being used in we 
                only want to display a portion of the graph data.
            In addition to only displaying relevant nodes, we want to add conditional
                attributes to topological_gen so we can visualize them in the tangled tree
                as well.
        Input:
            topological_gen (List[list]): Indicates layers of nodes.
            all_attributes (list): all attributes associated with a particular component.
            conditional_attributes (list): List of conditional attributes for a particular component
        Output:
            new_top_gen (List[list]): mimics structure of topological_gen but only
                includes the nodes we want
        '''

        pruned_topological_gen = []

        # For each layer(gen) in the topological generation list
        for i, layer in enumerate(topological_gen):
            
            current_layer = []
            next_layer = []
            
            # For each node in the layer
            for node in layer:
                
                # If the node is relevant to this component and is not a conditional attribute add it to the current layer.
                if node in all_attributes and node not in conditional_attributes:
                    current_layer.append(node)
                
                # If its a conditional attribute add it to a followup layer.
                if node in conditional_attributes:
                    next_layer.append(node)

            # Added layers to new pruned_topological_gen list
            if current_layer:
                pruned_topological_gen.append(current_layer)
            if next_layer:
                pruned_topological_gen.append(next_layer)

        return pruned_topological_gen

    def get_base_layers(self, topological_gen, child_parents, source_nodes, cn):
        '''
        Purpose:
            Reconfigure topological gen to move things back appropriate layers if 
            they would have a back reference.

            The Tangle Tree figure requrires an acyclic directed graph that has additional 
                layering rules between connected nodes.
                - If there is a backward connection then the line connecting them will
                    break (this would suggest a cyclic connection.)
                - Additionally if two or more nodes are connecting to a downstream node it is 
                    best to put both parent nodes at the same level, if possible, to 
                    prevent line breaks.
                - Also want to move any children nodes one layer below 
                    the parent node(s). If there are multiple parents, put one layer below the
                    parent that is furthest from the origin.

            This is an iterative process that needs to run twice to move all the nodes to their
            appropriate positions.
        Input:
            topological_gen: list of lists. Indicates layers of nodes.
            child_parents (dict):
                key: child
                value: list of the childs parents
            source_nodes: list, list of nodes that do not have a parent.
            cn: str, component name, default=''
        Output:
            base_layers: dict, key: component name, value: layer
                represents initial layering of toplogical_gen
            base_layers_copy_copy: dict, key: component name, value: layer
                represents the final layering after moving the components/attributes to
                their desired layer.c
        '''
        # Convert topological_gen to a dictionary
        base_layers = {com:i for i, lev in enumerate(topological_gen)
                                for com in lev}
        
        # Make another version to iterate on -- Cant set to equal or will overwrite the original.
        base_layers_copy = {com:i for i, lev in enumerate(topological_gen)
                                for com in lev}

        # Move child nodes one node downstream of their parents.
        for level in topological_gen:
            for node in level:
                
                # Check if node has a parent.
                if node in child_parents.keys():
                    
                    #node_level = base_layers[node]
                    # Look at the parents for the node.
                    parent_levels = []
                    for par in child_parents[node]:
                        
                        # Get the layer the parent is located at.
                        parent_levels.append(base_layers[par])
                        
                        # Get the max layer a parent of the node can be found.
                        max_parent_level = max(parent_levels)

                        # Move the node one layer beyond the max parent node position, so it will be downstream of its parents.
                        base_layers_copy[node] = max_parent_level + 1
        
        # Make another version of updated positions iterate on further.
        base_layers_copy_copy = base_layers_copy

        # Move parental source nodes if necessary.
        for level in topological_gen:
            for node in level:
                
                # Check if node has any parents.
                if node in child_parents.keys():
                    parent_levels = []
                    modify_par = []
                    
                    # For each parent get their position.
                    for par in child_parents[node]:
                        parent_levels.append(base_layers_copy[par])
                    
                    # If one of the parents is a source node move 
                    # it to the same level as the other nodes the child connects to so
                    # that the connections will not be backwards (and result in a broken line)
                    for par in child_parents[node]:
                        
                        # For a given parent determine if its a source node and that the parents 
                        # are not already at level 0, and the parent is not the current component node.
                        if (par in source_nodes and 
                            (parent_levels.count(parent_levels[0]) != len(parent_levels))
                            and par != cn):

                            # If so, remove its position from parent_levels
                            parent_levels.remove(base_layers_copy[par])
                            
                            # Add this parent to a list of parental positions to modify later.
                            modify_par.append(par)
                        
                        # Get the new max parent level for this node.
                        max_parent_level = max(parent_levels)
                        
                        # Move the node one position downstream of its max parent level.
                        base_layers_copy_copy[node] = max_parent_level + 1
                    
                    # For each parental position to modify, move the parents level up to the max_parent_level.
                    for par in modify_par:
                        base_layers_copy_copy[par] = max_parent_level
        
        return base_layers, base_layers_copy_copy

    def adjust_node_placement(self, base_layers_copy_copy, base_layers, topological_gen):
        '''Reorder nodes within topological_generations to match how they were ordered in base_layers_copy_copy
        Input:
            topological_gen: list of lists. Indicates layers of nodes.
            base_layers: dict, key: component name, value: layer
                represents initial layering of toplogical_gen
            base_layers_copy_copy: dict, key: component name, value: layer
                represents the final layering after moving the components/attributes to
                their desired layer.
        Output:
            topological_gen: same format but as the incoming topologial_gen but
                ordered to match base_layers_copy_copy.
        '''
        if self.figure_type == 'component':
            # For each node get its new layer in the tangled tree
            for node, i in base_layers_copy_copy.items():
                
                # Check if node is not already in the proper layer
                if node not in topological_gen[i]:

                    # If not put it in the appropriate layer
                    topological_gen[i].append(node)
                    
                    # Remove from inappropriate layer.
                    topological_gen[base_layers[node]].remove(node)
        
        elif self.figure_type == 'dependency':
            for node, i in base_layers_copy_copy.items():
                
                # Check if the location of the node is more than the number of 
                # layers topological gen current handles
                if i > len(topological_gen) - 1:

                    # If so, add node to new node at the end of topological_gen
                    topological_gen.append([node])
                    
                    # Remove the node from its previous position.
                    topological_gen[base_layers[node]].remove(node)
                
                # Else, check if node is not already in the proper layer
                elif node not in topological_gen[i]:

                    # If not put it in the appropriate layer
                    topological_gen[i].append(node)
                    
                    # Remove from inappropriate layer.
                    topological_gen[base_layers[node]].remove(node)
        return topological_gen

    def move_source_nodes_to_bottom_of_layer(self, node_layers, source_nodes):
        '''For aesthetic purposes move source nodes to the bottom of their respective layers.
        Input:
            node_layers (List(list)): Lists of lists of each layer and the nodes contained in that layer as strings.
            source_nodes (list): list of nodes that do not have a parent.
        Output:
            node_layers (List(list)): modified to move source nodes to the bottom of each layer.
        '''
        for i, layer in enumerate(node_layers):
            nodes_to_move = []
            for node in layer:
                if node in source_nodes:
                    nodes_to_move.append(node)
            for node in nodes_to_move:
                node_layers[i].remove(node)
                node_layers[i].append(node)
        return node_layers

    def get_layers_dict_list(self, node_layers, child_parents, parent_children, all_parent_children):
        '''Convert node_layers to a list of lists of dictionaries that specifies each node and its parents (if applicable).
        Inputs:
            node_layers: list of lists of each layer and the nodes contained in that layer as strings.
            child_parents (dict):
                key: child
                value: list of the childs parents
            parent_children (dict):
                key: parent
                value: list of the parents children
        Outputs:
            layers_list (List(list): list of lists of dictionaries that specifies each node and its parents (if applicable)
        '''
        num_layers = len(node_layers)
        layers_list = [[] for i in range(0, num_layers)]
        for i, layer in enumerate(node_layers):
            for node in layer:
                if node in child_parents.keys():
                    parents = child_parents[node]
                else: 
                    parents = []

                if node in parent_children.keys():
                    direct_children = parent_children[node]
                else: 
                    direct_children = []

                if node in all_parent_children.keys():
                    all_children = all_parent_children[node]
                else: 
                    all_children = []
                layers_list[i].append({'id': node, 'parents': parents, 'direct_children': direct_children, 'children': all_children})

        return layers_list

    def get_node_layers_json(self, topological_gen, source_nodes, child_parents, parent_children, cn='', all_parent_children=None):
        '''Return all the layers of a single tangled tree as a JSON String.
        Inputs:
            topological_gen:list of lists. Indicates layers of nodes.
            source_nodes: list of nodes that do not have a parent.
            child_parents (dict):
                key: child
                value: list of the childs parents
            parent_children (dict):
                key: parent
                value: list of the parents children
            all_parent_children (dict):
                key: parent
                value: list of the parents children (including all downstream nodes). Default to an empty dictionary
        Outputs:
            layers_json (JSON String): Layers of nodes in the tangled tree as a json string.
        '''

        base_layers, base_layers_copy_copy = self.get_base_layers(topological_gen, 
                child_parents, source_nodes, cn)

        # Rearrange node_layers to follow the pattern laid out in component layers.
        node_layers = self.adjust_node_placement(base_layers_copy_copy, 
            base_layers, topological_gen)

        # Move source nodes to the bottom of each layer.
        node_layers = self.move_source_nodes_to_bottom_of_layer(node_layers, source_nodes)

        # Convert layers to a list of dictionaries
        if not all_parent_children:
            # default to an empty dictionary 
            all_parent_children = dict()
            
        layers_dicts = self.get_layers_dict_list(node_layers, child_parents, parent_children, all_parent_children)

        # Convert dictionary to a JSON string
        layers_json = json.dumps(layers_dicts)

        return layers_json

    def save_outputs(self, save_file, layers_json, cn='', all_layers=[]):
        '''
        Inputs:
            save_file (bool): Indicates whether to save a file locally or not.:
            layers_json (JSON String): Layers of nodes in the tangled tree as a json string.
            cn (str): component name, default=''
            all_layers (list of json strings): Each string represents contains the layers for a single tangled tree.
                If a dependency figure the list is added to each time this function is called, so starts incomplete.
                default=[].
        Outputs:
            all_layers (list of json strings): 
                If save_file == False: Each string represents contains the layers for a single tangled tree.
                If save_file ==True: is an empty list.
        '''
        if save_file == True:
            if cn:
                output_file_name = f"{self.schema_abbr}_{self.figure_type}_{cn}_tangled_tree.json"
            else:
                output_file_name = f"{self.schema_abbr}_{self.figure_type}_tangled_tree.json"
            with open(os.path.join(self.json_output_path, output_file_name), 'w') as outfile:
                outfile.write(layers_json)
            logger.info(f"Tangled Tree JSON String saved to {os.path.join(self.json_output_path, output_file_name)}.")
            all_layers = layers_json
        elif save_file == False:
            all_layers.append(layers_json)
        return all_layers

    def get_ancestors_nodes(self, subgraph, components):
        """
        Inputs: 
            subgraph: networkX graph object
            components: a list of nodes 
        outputs: 
            all_parent_children: a dictionary that indicates a list of children (including all the intermediate children) of a given node
        """
        all_parent_children = {}
        for component in components: 
            all_ancestors = self.sg.se.get_nodes_ancestors(subgraph, component)
            all_parent_children[component] = all_ancestors

        return all_parent_children

    def get_tangled_tree_layers(self, save_file=True):
        '''Based on user indicated figure type, construct the layers of nodes of a tangled tree.
        Inputs:
            save_file (bool): Indicates whether to save a file locally or not.
        Outputs:
            all_layers (list of json strings): 
                If save_file == False: Each string represents contains the layers for a single tangled tree.
                If save_file ==True: is an empty list.

        Note on Dependency Tangled Tree:
            If there are many conditional requirements associated with a depependency, and those
            conditional requirements have overlapping attributes associated with them
            the tangled tree will only report one
        
        '''
        # Gather the data model's, topological generations, nodes and edges
        topological_gen, nodes, edges, subg = self.get_topological_generations()

        if self.figure_type == 'component':
            # Gather all source nodes
            source_nodes = self.find_source_nodes(nodes, edges)
            
            # Map all children to their parents and vice versa
            child_parents, parent_children = self.get_parent_child_dictionary(nodes, edges)

            # find all the downstream nodes
            all_parent_children = self.get_ancestors_nodes(subg, parent_children.keys())
            
            # Get the layers that each node belongs to.
            layers_json = self.get_node_layers_json(topological_gen, source_nodes, child_parents, parent_children, all_parent_children=all_parent_children)

            # If indicated save outputs locally else gather all layers.
            all_layers = self.save_outputs(save_file, layers_json)         

        if self.figure_type == 'dependency':
            # Get component digraph and nodes.
            component_dg = self.sg.se.get_digraph_by_edge_type('requiresComponent')
            component_nodes = component_dg.nodes()

            # Get table of attributes.
            attributes_csv_str = self.ae.parse_attributes(save_file=False)
            attributes_df = pd.read_table(StringIO(attributes_csv_str), sep=",")

            
            all_layers =[]
            for cn in component_nodes:
                # Gather attribute and dependency information per node
                conditional_attributes, ca_alias, all_attributes = self.gather_component_dependency_info(cn, attributes_df)

                # Gather all source nodes
                source_nodes = self.find_source_nodes(component_nodes, edges, all_attributes)

                # Alias the conditional requirement edge back to its actual parent label,
                # then apply aliasing back to the edges
                aliased_edges = self.alias_edges(ca_alias, edges)

                # Gather relationships between children and their parents.
                child_parents, parent_children = self.get_parent_child_dictionary(nodes,
                    aliased_edges, all_attributes)

                # Remake topological_gen so it has only relevant nodes.
                pruned_topological_gen = self.prune_expand_topological_gen(topological_gen, all_attributes, conditional_attributes)

                # Get the layers that each node belongs to.
                layers_json = self.get_node_layers_json(pruned_topological_gen, source_nodes, child_parents, parent_children, cn)

                # If indicated save outputs locally else, gather all layers.
                all_layers = self.save_outputs(save_file, layers_json, cn, all_layers)
        return all_layers

   


from copy import deepcopy
import os
import uuid  # used to generate unique names for entities
import json
import atexit
import logging
import secrets

# allows specifying explicit variable types
from typing import Dict, List, Tuple, Sequence, Union
from collections import OrderedDict

import numpy as np
import pandas as pd
import re
import synapseclient
from time import sleep

from synapseclient import (
    Synapse,
    File,
    Folder,
    Table,
    Schema,
    EntityViewSchema,
    EntityViewType,
    Column,
    as_table_columns,
)

from synapseclient.table import CsvFileTable
from synapseclient.table import build_table
from synapseclient.annotations import from_synapse_annotations
from synapseclient.core.exceptions import SynapseHTTPError, SynapseAuthenticationError, SynapseUnmetAccessRestrictions
from synapseutils import walk
from synapseutils.copy_functions import changeFileMetaData

import uuid

from schematic.utils.df_utils import update_df, load_df
from schematic.utils.validate_utils import comma_separated_list_regex, rule_in_rule_list
from schematic.schemas.explorer import SchemaExplorer
from schematic.schemas.generator import SchemaGenerator
from schematic.store.base import BaseStorage
from schematic.exceptions import MissingConfigValueError, AccessCredentialsError

from schematic import CONFIG

logger = logging.getLogger(__name__)

class SynapseStorage(BaseStorage):
    """Implementation of Storage interface for datasets/files stored on Synapse.
    Provides utilities to list files in a specific project; update files annotations, create fileviews, etc.

    TODO: Need to define the interface and rename and/or refactor some of the methods below.
    """

    def __init__(
        self,
        token: str = None,  # optional parameter retrieved from browser cookie
        access_token: str = None,
        input_token: str = None,
        project_scope: List = None,
    ) -> None:
        """Initializes a SynapseStorage object.
        Args:
            syn: an object of type synapseclient.
            token: optional token parameter (typically a 'str') as found in browser cookie upon login to synapse.
            access_token: optional access token (personal or oauth)
            TODO: move away from specific project setup and work with an interface that Synapse specifies (e.g. based on schemas).
        Exceptions:
            KeyError: when the 'storage' config object is missing values for essential keys.
            AttributeError: when the 'storageFileview' attribute (of class SynapseStorage) does not have a value associated with it.
            synapseclient.core.exceptions.SynapseHTTPError: check if the current user has permission to access the Synapse entity.
            ValueError: when Admin fileview cannot be found (describe further).
        Typical usage example:
            syn_store = SynapseStorage()
        """

        self.syn = self.login(token, access_token, input_token)
        self.project_scope = project_scope


        # check if "master_fileview" has been set
        try: 
            self.storageFileview = CONFIG["synapse"]["master_fileview"]
        except KeyError: 
            raise MissingConfigValueError(("synapse", "master_fileview"))

        # check if "manifest_basename" has been set
        try: 
            self.manifest = CONFIG["synapse"]["manifest_basename"]
        except KeyError: 
            raise MissingConfigValueError(("synapse", "manifest_basename"))

        try:
            self.storageFileview = CONFIG["synapse"]["master_fileview"]
            self.manifest = CONFIG["synapse"]["manifest_basename"]
            if self.project_scope:
                self.storageFileviewTable = self.syn.tableQuery(
                    f"SELECT * FROM {self.storageFileview} WHERE projectId IN {tuple(self.project_scope + [''])}"
                ).asDataFrame()
            else:
                # get data in administrative fileview for this pipeline
                self.storageFileviewTable = self.syn.tableQuery(
                    "SELECT * FROM " + self.storageFileview
                ).asDataFrame()
        except AttributeError:
            raise AttributeError("storageFileview attribute has not been set.")
        except SynapseHTTPError:
            raise AccessCredentialsError(self.storageFileview)

    @staticmethod
    def login(token=None, access_token=None, input_token=None):
        # If no token is provided, try retrieving access token from environment
        if not token and not access_token and not input_token:
            access_token = os.getenv("SYNAPSE_ACCESS_TOKEN")

        # login using a token
        if token:
            syn = synapseclient.Synapse()

            try:
                syn.login(sessionToken=token, silent=True)
            except synapseclient.core.exceptions.SynapseHTTPError:
                raise ValueError("Please make sure you are logged into synapse.org.")
        elif access_token:
            syn = synapseclient.Synapse()
            syn.default_headers["Authorization"] = f"Bearer {access_token}"
        elif input_token: 
            try: 
                syn = synapseclient.Synapse()
                syn.default_headers["Authorization"] = f"Bearer {input_token}"
            except synapseclient.core.exceptions.SynapseHTTPError:
                raise ValueError("No access to resources. Please make sure that your token is correct")
        else:
            # login using synapse credentials provided by user in .synapseConfig (default) file
            syn = synapseclient.Synapse(configPath=CONFIG.SYNAPSE_CONFIG_PATH)
            syn.login(silent=True)
            
        return syn

    def missing_entity_handler(method):
        def wrapper(*args, **kwargs):
            try:
                return method(*args, **kwargs)
            except(SynapseHTTPError) as ex:
                str_message = str(ex).replace("\n","")
                if 'trash' in str_message or 'does not exist' in str_message:
                    logging.warning(str_message)
                    return None
                else:
                    raise ex
        return wrapper


    def getStorageFileviewTable(self):
        """ Returns the storageFileviewTable obtained during initialization.
        """
        return self.storageFileviewTable

    def getPaginatedRestResults(self, currentUserId: str) -> Dict[str, str]:
        """Gets the paginated results of the REST call to Synapse to check what projects the current user has access to.

        Args:
            currentUserId: synapse id for the user whose projects we want to get.

        Returns:
            A dictionary with a next page token and the results.
        """
        all_results = self.syn.restGET(
            "/projects/user/{principalId}".format(principalId=currentUserId)
        )

        while (
            "nextPageToken" in all_results
        ):  # iterate over next page token in results while there is any
            results_token = self.syn.restGET(
                "/projects/user/{principalId}?nextPageToken={nextPageToken}".format(
                    principalId=currentUserId,
                    nextPageToken=all_results["nextPageToken"],
                )
            )
            all_results["results"].extend(results_token["results"])

            if "nextPageToken" in results_token:
                all_results["nextPageToken"] = results_token["nextPageToken"]
            else:
                del all_results["nextPageToken"]

        return all_results

    def getStorageProjects(self, project_scope: List = None) -> List[str]:
        """Gets all storage projects the current user has access to, within the scope of the 'storageFileview' attribute.

        Returns:
            A list of storage projects the current user has access to; the list consists of tuples (projectId, projectName).
        """

        # get the set of all storage Synapse project accessible for this pipeline
        storageProjects = self.storageFileviewTable["projectId"].unique()

        # get the set of storage Synapse project accessible for this user

        # get current user name and user ID
        currentUser = self.syn.getUserProfile()
        currentUserName = currentUser.userName
        currentUserId = currentUser.ownerId

        # get a list of projects from Synapse
        currentUserProjects = self.getPaginatedRestResults(currentUserId)

        # prune results json filtering project id
        currentUserProjects = [
            currentUserProject.get("id")
            for currentUserProject in currentUserProjects["results"]
        ]

        # find set of user projects that are also in this pipeline's storage projects set
        storageProjects = list(set(storageProjects) & set(currentUserProjects))
        
        # Limit projects to scope if specified
        if project_scope:
            storageProjects = list(set(storageProjects) & set(project_scope))

            if not storageProjects:
                raise Warning(
                    f"There are no projects that the user has access to that match the criteria of the specified project scope: {project_scope}"
                )
        
        # prepare a return list of project IDs and names
        projects = []
        for projectId in storageProjects:
            projectName = self.syn.get(projectId, downloadFile=False).name
            projects.append((projectId, projectName))

        sorted_projects_list = sorted(projects, key=lambda tup: tup[0])

        return sorted_projects_list

    def getStorageDatasetsInProject(self, projectId: str) -> List[str]:
        """Gets all datasets in folder under a given storage project that the current user has access to.

        Args:
            projectId: synapse ID of a storage project.

        Returns:
            A list of datasets within the given storage project; the list consists of tuples (datasetId, datasetName).
            None: If the projectId cannot be found on Synapse.
        """

        # select all folders and fetch their names from within the storage project;
        # if folder content type is defined, only select folders that contain datasets
        areDatasets = False
        if "contentType" in self.storageFileviewTable.columns:
            foldersTable = self.storageFileviewTable[
                (self.storageFileviewTable["contentType"] == "dataset")
                & (self.storageFileviewTable["projectId"] == projectId)
            ]
            areDatasets = True
        else:
            foldersTable = self.storageFileviewTable[
                (self.storageFileviewTable["type"] == "folder")
                & (self.storageFileviewTable["parentId"] == projectId)
            ]

        # get an array of tuples (folderId, folderName)
        # some folders are part of datasets; others contain datasets
        # each dataset parent is the project; folders part of a dataset have another folder as a parent
        # to get folders if and only if they contain datasets for each folder
        # check if folder's parent is the project; if so that folder contains a dataset,
        # unless the folder list has already been filtered to dataset folders based on contentType attribute above

        datasetList = []
        folderProperties = ["id", "name"]
        for folder in list(
            foldersTable[folderProperties].itertuples(index=False, name=None)
        ):
            datasetList.append(folder)

        sorted_dataset_list = sorted(datasetList, key=lambda tup: tup[0])

        return sorted_dataset_list

    def getFilesInStorageDataset(
        self, datasetId: str, fileNames: List = None, fullpath: bool = True
    ) -> List[Tuple[str, str]]:
        """Gets all files in a given dataset folder.

        Args:
            datasetId: synapse ID of a storage dataset.
            fileNames: get a list of files with particular names; defaults to None in which case all dataset files are returned (except bookkeeping files, e.g.
            metadata manifests); if fileNames is not None, all files matching the names in the fileNames list are returned if present.
            fullpath: if True return the full path as part of this filename; otherwise return just base filename

        Returns:
            A list of files; the list consists of tuples (fileId, fileName).

        Raises:
            ValueError: Dataset ID not found.
        """

        # select all files within a given storage dataset folder (top level folder in a Synapse storage project or folder marked with contentType = 'dataset')
        walked_path = walk(self.syn, datasetId)

        file_list = []

        # iterate over all results
        for dirpath, dirname, filenames in walked_path:

            # iterate over all files in a folder
            for filename in filenames:

                if (not "manifest" in filename[0] and not fileNames) or (
                    fileNames and filename[0] in fileNames
                ):

                    # don't add manifest to list of files unless it is specified in the list of specified fileNames; return all found files
                    # except the manifest if no fileNames have been specified
                    # TODO: refactor for clarity/maintainability

                    if fullpath:
                        # append directory path to filename
                        filename = (dirpath[0] + "/" + filename[0], filename[1])

                    # add file name file id tuple, rearranged so that id is first and name follows
                    file_list.append(filename[::-1])

        return file_list

    def getDatasetManifest(
        self, datasetId: str, downloadFile: bool = False, newManifestName: str='',
    ) -> List[str]:
        """Gets the manifest associated with a given dataset.

        Args:
            datasetId: synapse ID of a storage dataset.
            downloadFile: boolean argument indicating if manifest file in dataset should be downloaded or not.

        Returns:
            manifest_syn_id (String): Synapse ID of exisiting manifest file.
            manifest_data (synapseclient.entity.File): Synapse entity if downloadFile is True.
            "" (String): No pre-exisiting manifest in dataset.
        """

        # get a list of files containing the manifest for this dataset (if any)
        all_files = self.storageFileviewTable

        manifest_re=re.compile(os.path.basename(self.manifest)+".*.[tc]sv")
        manifest = all_files[
            (all_files['name'].str.contains(manifest_re,regex=True))
            & (all_files["parentId"] == datasetId)
        ]

        manifest = manifest[["id", "name"]]
        censored_regex=re.compile('.*censored.*')
        
        # if there is no pre-exisiting manifest in the specified dataset
        if manifest.empty:
            return ""

        # if there is an exisiting manifest
        else:
            # retrieve data from synapse

            # if a censored manifest exists for this dataset
            censored = manifest['name'].str.contains(censored_regex)
            if any(censored):
                # Try to use uncensored manifest first
                not_censored=~censored
                if any(not_censored):
                    manifest_syn_id=manifest[not_censored]["id"][0]

            #otherwise, use the first (implied only) version that exists
            else:
                manifest_syn_id = manifest["id"][0]


            # if the downloadFile option is set to True
            if downloadFile:
                # enables retrying if user does not have access to uncensored manifest
                while True:
                    # pass synID to synapseclient.Synapse.get() method to download (and overwrite) file to a location
                    try:
                        if 'manifest_folder' in CONFIG['synapse'].keys():
                            manifest_data = self.syn.get(
                                manifest_syn_id,
                                downloadLocation=CONFIG["synapse"]["manifest_folder"],
                                ifcollision="overwrite.local",
                            )
                            break
                        # if no manifest folder is set, download to cache
                        else:
                            manifest_data = self.syn.get(
                                manifest_syn_id,
                            )
                            break
                    # If user does not have access to uncensored manifest, use censored instead
                    except(SynapseUnmetAccessRestrictions):
                            manifest_syn_id=manifest[censored]["id"][0]
                    
                # Rename manifest file if indicated by user.
                if newManifestName:
                    if os.path.exists(manifest_data['path']):
                        # Rename the file we just made to the new name
                        new_manifest_filename = newManifestName + '.csv'
                        new_manifest_path_name = manifest_data['path'].replace(manifest['name'][0], new_manifest_filename)
                        os.rename(manifest_data['path'], new_manifest_path_name)

                        # Update file names/paths in manifest_data
                        manifest_data['name'] = new_manifest_filename
                        manifest_data['filename'] = new_manifest_filename
                        manifest_data['path'] = new_manifest_path_name

                return manifest_data


            return manifest_syn_id

    def getDataTypeFromManifest(self, manifestId:str):
        """Fetch a manifest and return data types of all columns
        Args: 
            manifestId: synapse ID of a manifest
        """
        # get manifest file path 
        manifest_filepath = self.syn.get(manifestId).path

        # load manifest dataframe 
        manifest = load_df(manifest_filepath, preserve_raw_input=False, data_model=False)

        # convert the dataFrame to use best possible dtypes.
        manifest_new = manifest.convert_dtypes()

        # get data types of columns
        result = manifest_new.dtypes.to_frame('dtypes').reset_index()

        # return the result as a dictionary 
        result_dict = result.set_index('index')['dtypes'].astype(str).to_dict()


        return result_dict

    def updateDatasetManifestFiles(self, sg: SchemaGenerator, datasetId: str, store:bool = True) -> Union[Tuple[str, pd.DataFrame], None]:
        """Fetch the names and entity IDs of all current files in dataset in store, if any; update dataset's manifest with new files, if any.

        Args:
            datasetId: synapse ID of a storage dataset.
            store: if set to True store updated manifest in asset store; if set to False
            return a Pandas dataframe containing updated manifest but do not store to asset store


        Returns:
            Synapse ID of updated manifest and Pandas dataframe containing the updated manifest. 
            If there is no existing manifest return None
        """

        # get existing manifest Synapse ID
        manifest_id = self.getDatasetManifest(datasetId)

        # if there is no manifest return None
        if not manifest_id:
            return None

        manifest_filepath = self.syn.get(manifest_id).path
        manifest = load_df(manifest_filepath)

        # get current list of files
        dataset_files = self.getFilesInStorageDataset(datasetId)

        # update manifest with additional filenames, if any
        # note that if there is an existing manifest and there are files in the dataset
        # the columns Filename and entityId are assumed to be present in manifest schema
        # TODO: use idiomatic panda syntax
        if dataset_files:
            new_files = {"Filename": [], "entityId": []}

            # find new files if any
            for file_id, file_name in dataset_files:
                if not file_id in manifest["entityId"].values:
                    new_files["Filename"].append(file_name)
                    new_files["entityId"].append(file_id)

            # update manifest so that it contain new files
            new_files = pd.DataFrame(new_files)
            manifest = (
                pd.concat([manifest, new_files], sort=False)
                .reset_index()
                .drop("index", axis=1)
            )

            # update the manifest file, so that it contains the relevant entity IDs
            if store:
                manifest.to_csv(manifest_filepath, index=False)

                # store manifest and update associated metadata with manifest on Synapse
                manifest_id = self.associateMetadataWithFiles(sg, manifest_filepath, datasetId)

        manifest = manifest.fillna("") 
        
        return manifest_id, manifest

    def getProjectManifests(self, projectId: str) -> List[str]:
        """Gets all metadata manifest files across all datasets in a specified project.

        Returns: A list of datasets per project; metadata manifest Synapse ID for each dataset; and the corresponding schema component of the manifest
                 as a list of tuples, one for each manifest:
                    [
                        (
                            (datasetId, dataName),
                            (manifestId, manifestName),
                            (componentSchemaLabel, componentSchemaLabel) TODO: # get component name from schema
                        ),
                        ...
                    ]

        TODO: Return manifest URI instead of Synapse ID for interoperability with other implementations of a store interface
        """
        component=None
        entity=None
        manifests = []

        datasets = self.getStorageDatasetsInProject(projectId)

        for (datasetId, datasetName) in datasets:
            # encode information about the manifest in a simple list (so that R clients can unpack it)
            # eventually can serialize differently
                
            # Get synID of manifest for a dataset
            manifestId = self.getDatasetManifest(datasetId)

            # If a manifest exists, get the annotations for it, else return base 'manifest' tuple
            if manifestId:
                annotations = self.getFileAnnotations(manifestId)

                # If manifest has annotations specifying component, use that
                if 'Component' in annotations:
                    component = annotations['Component']
                    entity = self.syn.get(manifestId, downloadFile=False)
                    manifest_name = entity["properties"]["name"]

                # otherwise download the manifest and parse for information
                elif 'Component' not in annotations or not annotations:
                    logging.debug(
                        f"No component annotations have been found for manifest {manifestId}. "
                        "The manifest will be downloaded and parsed instead. "
                        "For increased speed, add component annotations to manifest."
                        )

                    manifest_info = self.getDatasetManifest(datasetId,downloadFile=True)
                    manifest_name = manifest_info["properties"]["name"]
                    manifest_path = manifest_info["path"]

                    manifest_df = load_df(manifest_path)

                    # Get component from component column if it exists
                    if "Component" in manifest_df and not manifest_df["Component"].empty:
                        list(set(manifest_df['Component']))
                        component = list(set(manifest_df["Component"]))

                        #Added to address issues raised during DCA testing
                        if '' in component:
                            component.remove('')

                        if len(component) == 1:
                            component = component[0]
                        elif len(component) > 1:
                            logging.warning(
                            f"Manifest {manifestId} is composed of multiple components. Schematic does not support mulit-component manifests at this time."
                            "Behavior of manifests with multiple components is undefined"
                            )              

            #save manifest list with applicable informaiton
            if component:
                manifest = (
                    (datasetId, datasetName),
                    (manifestId, manifest_name),
                    (component, component),
                )
            elif manifestId:
                logging.debug(f"Manifest {manifestId} does not have an associated Component")
                manifest = (
                    (datasetId, datasetName),
                    (manifestId, manifest_name),
                    ("", ""),
                )
            else:
                manifest = (
                    (datasetId, datasetName),
                    ("", ""),
                    ("", ""),
                )

            if manifest:
                manifests.append(manifest)
                
        return manifests

    def upload_project_manifests_to_synapse(self, projectId: str) -> List[str]:
        """Upload all metadata manifest files across all datasets in a specified project as tables in Synapse.

        Returns: String of all the manifest_table_ids of all the manifests that have been loaded.
        """

        manifests = []
        manifest_loaded = []
        datasets = self.getStorageDatasetsInProject(projectId)

        for (datasetId, datasetName) in datasets:
            # encode information about the manifest in a simple list (so that R clients can unpack it)
            # eventually can serialize differently

            manifest = ((datasetId, datasetName), ("", ""), ("", ""))

            manifest_info = self.getDatasetManifest(datasetId, downloadFile=True)
            if manifest_info:
                manifest_id = manifest_info["properties"]["id"]
                manifest_name = manifest_info["properties"]["name"]
                manifest_path = manifest_info["path"]
                manifest_df = load_df(manifest_path)
                manifest_table_id = upload_format_manifest_table(manifest, datasetId, datasetName)
                manifest_loaded.append(datasetName)
        return manifest_loaded

    def upload_annotated_project_manifests_to_synapse(self, projectId:str, path_to_json_ld: str, dry_run: bool = False) -> List[str]:
        '''
        Purpose:
            For all manifests in a project, upload them as a table and add annotations manifest csv.
            Assumes the manifest is already present as a CSV in a dataset in the project.

        '''

        sg = SchemaGenerator(path_to_json_ld)
        manifests = []
        manifest_loaded = []
        datasets = self.getStorageDatasetsInProject(projectId)
        for (datasetId, datasetName) in datasets:
            # encode information about the manifest in a simple list (so that R clients can unpack it)
            # eventually can serialize differently

            manifest = ((datasetId, datasetName), ("", ""), ("", ""))
            manifests.append(manifest)

            manifest_info = self.getDatasetManifest(datasetId, downloadFile=True)

            if manifest_info:
                manifest_id = manifest_info["properties"]["id"]
                manifest_name = manifest_info["properties"]["name"]
                manifest_path = manifest_info["path"]
                manifest = ((datasetId, datasetName), (manifest_id, manifest_name), ("", ""))
                if not dry_run:
                    manifest_syn_id = self.associateMetadataWithFiles(sg, manifest_path, datasetId, manifest_record_type='table')
                manifest_loaded.append(manifest)
            
        return manifests, manifest_loaded


    def move_entities_to_new_project(self, projectId: str, newProjectId: str, returnEntities: bool = False, dry_run: bool = False):
        """
        For each manifest csv in a project, look for all the entitiy ids that are associated.
        Look up the entitiy in the files, move the entity to new project.
        """

        manifests = []
        manifest_loaded = []
        datasets = self.getStorageDatasetsInProject(projectId)
        if datasets:
            for (datasetId, datasetName) in datasets:
                # encode information about the manifest in a simple list (so that R clients can unpack it)
                # eventually can serialize differently

                manifest = ((datasetId, datasetName), ("", ""), ("", ""))
                manifests.append(manifest)

                manifest_info = self.getDatasetManifest(datasetId, downloadFile=True)
                if manifest_info:
                    manifest_id = manifest_info["properties"]["id"]
                    manifest_name = manifest_info["properties"]["name"]
                    manifest_path = manifest_info["path"]
                    manifest_df = load_df(manifest_path)

                    manifest = ((datasetId, datasetName), (manifest_id, manifest_name), ("", ""))
                    manifest_loaded.append(manifest)

                    annotation_entities = self.storageFileviewTable[
                            (self.storageFileviewTable['id'].isin(manifest_df['entityId']))
                            & (self.storageFileviewTable['type'] == 'folder')
                        ]['id']

                    if returnEntities:
                        for entityId in annotation_entities: 
                            if not dry_run:
                                self.syn.move(entityId, datasetId)
                            else:
                                logging.info(f"{entityId} will be moved to folder {datasetId}.")
                    else:                
                        # generate project folder
                        archive_project_folder = Folder(projectId+'_archive', parent = newProjectId)
                        archive_project_folder = self.syn.store(archive_project_folder)
        
                        # generate dataset folder
                        dataset_archive_folder = Folder("_".join([datasetId,datasetName,'archive']), parent = archive_project_folder.id)
                        dataset_archive_folder = self.syn.store(dataset_archive_folder)                    

                        for entityId in annotation_entities:
                            # move entities to folder
                            if not dry_run:
                                self.syn.move(entityId, dataset_archive_folder.id)
                            else:
                                logging.info(f"{entityId} will be moved to folder {dataset_archive_folder.id}.")
        else:
            raise LookupError(
                f"No datasets were found in the specified project: {projectId}. Re-check specified master_fileview in CONFIG and retry."
            )
        return manifests, manifest_loaded

    def get_synapse_table(self, synapse_id: str) -> Tuple[pd.DataFrame, CsvFileTable]:
        """Download synapse table as a pd dataframe; return table schema and etags as results too

        Args:
            synapse_id: synapse ID of the table to query
        """

        results = self.syn.tableQuery("SELECT * FROM {}".format(synapse_id))
        df = results.asDataFrame(rowIdAndVersionInIndex=False)

        return df, results

    def _get_tables(self, datasetId: str = None) -> List[Table]:
        project = self.syn.get(self.getDatasetProject(datasetId))
        return list(self.syn.getChildren(project, includeTypes=["table"]))

    def get_table_info(self, datasetId: str = None) -> List[str]:
        """Gets the names of the tables in the schema
        Returns:
            list[str]: A list of table names
        """
        tables = self._get_tables(datasetId)
        if tables:
            return {table["name"]: table["id"] for table in tables}
        else: 
            return {None:None}

    @missing_entity_handler
    def upload_format_manifest_table(self, se, manifest, datasetId, table_name, restrict, useSchemaLabel):
        # Rename the manifest columns to display names to match fileview
        table_info = self.get_table_info(datasetId)

        blacklist_chars = ['(', ')', '.', ' ', '-']
        manifest_columns = manifest.columns.tolist()

        table_manifest=deepcopy(manifest)

        if useSchemaLabel:
            cols = [
                se.get_class_label_from_display_name(
                    str(col)
                    ).translate({ord(x): '' for x in blacklist_chars})
                for col in manifest_columns
            ]

            cols = list(map(lambda x: x.replace('EntityId', 'entityId'), cols))


            # Reset column names in table manifest
            table_manifest.columns = cols

        #move entity id to end of df
        entity_col = table_manifest.pop('entityId')
        table_manifest.insert(len(table_manifest.columns), 'entityId', entity_col)

        # Get the column schema
        col_schema = as_table_columns(table_manifest)

        # Set uuid column length to 64 (for some reason not being auto set.)
        for i, col in enumerate(col_schema):
            if col['name'] == 'Uuid':
                col_schema[i]['maximumSize'] = 64

        # Put table manifest onto synapse
        schema = Schema(name=table_name, columns=col_schema, parent=self.getDatasetProject(datasetId))
        if table_name not in table_info.keys():
            manifest_table_id = self.make_synapse_table(table_to_load = table_manifest, 
                                    dataset_id = datasetId,
                                    existingTableId = None, 
                                    table_name = table_name, 
                                    column_type_dictionary = col_schema, 
                                    restrict = restrict,
                                    manipulation = 'replace')
        else:
            manifest_table_id = self.make_synapse_table(table_to_load = table_manifest, 
                                    dataset_id = datasetId,
                                    existingTableId = table_info[table_name], 
                                    table_name = table_name, 
                                    column_type_dictionary = col_schema,
                                    restrict = restrict,
                                    manipulation = 'replace')


        return manifest_table_id, manifest, table_manifest

    def uplodad_manifest_file(self, manifest, metadataManifestPath, datasetId, restrict_manifest, component_name = ''):
        # Update manifest to have the new entityId column
        manifest.to_csv(metadataManifestPath, index=False)

        # store manifest to Synapse as a CSV
        # update file name
        file_name_full = metadataManifestPath.split('/')[-1]
        file_extension = file_name_full.split('.')[-1]

        # Differentiate "censored" and "uncensored" manifest
        if "censored" in file_name_full: 
            file_name_new = os.path.basename(CONFIG["synapse"]["manifest_basename"]) + "_" + component_name + "_censored" + '.' + file_extension
        else: 
            file_name_new = os.path.basename(CONFIG["synapse"]["manifest_basename"]) + "_" + component_name + '.' + file_extension

        manifestSynapseFile = File(
            metadataManifestPath,
            description="Manifest for dataset " + datasetId,
            parent=datasetId,
            name=file_name_new
        )

        manifest_synapse_file_id = self.syn.store(manifestSynapseFile, isRestricted = restrict_manifest).id
        changeFileMetaData(syn = self.syn, entity = manifest_synapse_file_id, downloadAs = file_name_new)
        
        return manifest_synapse_file_id

    @missing_entity_handler
    def format_row_annotations(self, se, sg, row, entityId, useSchemaLabel, hideBlanks):
        # prepare metadata for Synapse storage (resolve display name into a name that Synapse annotations support (e.g no spaces, parenthesis)
        # note: the removal of special characters, will apply only to annotation keys; we are not altering the manifest
        # this could create a divergence between manifest column and annotations. this should be ok for most use cases.
        # columns with special characters are outside of the schema
        metadataSyn = {}
        blacklist_chars = ['(', ')', '.', ' ', '-']
        
        for k, v in row.to_dict().items():

            if useSchemaLabel:
                keySyn = se.get_class_label_from_display_name(str(k)).translate({ord(x): '' for x in blacklist_chars})
            else:
                keySyn = str(k)

            # Skip `Filename` and `ETag` columns when setting annotations
            if keySyn in ["Filename", "ETag", "eTag"]:
                continue

            # truncate annotation values to 500 characters if the
            # size of values is greater than equal to 500 characters
            # add an explicit [truncatedByDataCuratorApp] message at the end
            # of every truncated message to indicate that the cell value
            # has been truncated
            if isinstance(v, str) and len(v) >= 500:
                v = v[0:472] + "[truncatedByDataCuratorApp]"

            metadataSyn[keySyn] = v
        
        # set annotation(s) for the various objects/items in a dataset on Synapse
        annos = self.syn.get_annotations(entityId)
        csv_list_regex=comma_separated_list_regex()
        for anno_k, anno_v in metadataSyn.items():
            
            # Remove keys with nan or empty string values from dict of annotations to be uploaded
            # if present on current data annotation
            if hideBlanks and (anno_v == '' or (isinstance(anno_v,float) and np.isnan(anno_v))):
                annos.pop(anno_k) if anno_k in annos.keys() else annos
            # Otherwise save annotation as approrpriate
            else:
                if isinstance(anno_v,float) and np.isnan(anno_v):
                        annos[anno_k] = ""
                elif isinstance(anno_v,str) and re.fullmatch(csv_list_regex, anno_v) and rule_in_rule_list('list', sg.get_node_validation_rules(anno_k)):
                    annos[anno_k] = anno_v.split(",")
                else:
                    annos[anno_k] = anno_v
                
        return annos

    @missing_entity_handler
    def format_manifest_annotations(self, manifest, manifest_synapse_id):
        '''
        Set annotations for the manifest (as a whole) so they can be applied to the manifest table or csv.
        For now just getting the Component.
        '''
        
        entity = self.syn.get(manifest_synapse_id, downloadFile=False)
        is_file = entity.concreteType.endswith(".FileEntity")
        is_table = entity.concreteType.endswith(".TableEntity")

        if is_file:

            # Get file metadata
            metadata = self.getFileAnnotations(manifest_synapse_id)

            # If there is a defined component add it to the metadata.
            if 'Component' in manifest.columns:
                # Gather component information
                component = manifest['Component'].unique()
                
                # Double check that only a single component is listed, else raise an error.
                try:
                    len(component) == 1
                except ValueError as err:
                    raise ValueError(
                        f"Manifest has more than one component. Please check manifest and resubmit."
                    ) from err

                # Add component to metadata
                metadata['Component'] = component[0]
        
        elif is_table:
            # Get table metadata
            metadata = self.getTableAnnotations(manifest_synapse_id)
        
        # Get annotations
        annos = self.syn.get_annotations(manifest_synapse_id)

        # Add metadata to the annotations
        for annos_k, annos_v in metadata.items():
            annos[annos_k] = annos_v

        return annos
    '''
    def annotate_upload_manifest_table(self, manifest, datasetId, metadataManifestPath,
        useSchemaLabel: bool = True, hideBlanks: bool = False, restrict_manifest = False):
        """
        Purpose:
            Works very similarly to associateMetadataWithFiles except takes in the manifest
            rather than the manifest path

        """
        
        # Add uuid for table updates and fill.
        if not "Uuid" in manifest.columns:
            manifest["Uuid"] = ''

        for idx,row in manifest.iterrows():
            if not row["Uuid"]:
                gen_uuid = uuid.uuid4()
                row["Uuid"] = gen_uuid
                manifest.loc[idx, 'Uuid'] = gen_uuid

        # add entityId as a column if not already there or
        # fill any blanks with an empty string.
        if not "entityId" in manifest.columns:
            manifest["entityId"] = ""
        else:
            manifest["entityId"].fillna("", inplace=True)

        # get a schema explorer object to ensure schema attribute names used in manifest are translated to schema labels for synapse annotations
        se = SchemaExplorer()

        # Create table name here.
        if 'Component' in manifest.columns:
            table_name = manifest['Component'][0].lower() + '_synapse_storage_manifest_table'
        else:
            table_name = 'synapse_storage_manifest_table'

        # Upload manifest as a table and get the SynID and manifest
        manifest_synapse_table_id, manifest, table_manifest = self.upload_format_manifest_table(
                                                    se, manifest, datasetId, table_name, restrict = restrict_manifest, useSchemaLabel=useSchemaLabel,)
            
        # Iterate over manifest rows, create Synapse entities and store corresponding entity IDs in manifest if needed
        # also set metadata for each synapse entity as Synapse annotations
        for idx, row in manifest.iterrows():
            if not row["entityId"]:
                # If not using entityIds, fill with manifest_table_id so 
                row["entityId"] = manifest_synapse_table_id
                entityId = ''
            else:
                # get the entity id corresponding to this row
                entityId = row["entityId"]

        # Load manifest to synapse as a CSV File
        manifest_synapse_file_id = self.uplodad_manifest_file(manifest, metadataManifestPath, datasetId, restrict_manifest)
        
        # Get annotations for the file manifest.
        manifest_annotations = self.format_manifest_annotations(manifest, manifest_synapse_file_id)
        
        self.syn.set_annotations(manifest_annotations)

        logger.info("Associated manifest file with dataset on Synapse.")
        
        # Update manifest Synapse table with new entity id column.
        self.make_synapse_table(
            table_to_load = table_manifest,
            dataset_id = datasetId,
            existingTableId = manifest_synapse_table_id,
            table_name = table_name,
            update_col = 'Uuid',
            specify_schema = False,
            )
        
        # Get annotations for the table manifest
        manifest_annotations = self.format_manifest_annotations(manifest, manifest_synapse_table_id)
        self.syn.set_annotations(manifest_annotations)
        return manifest_synapse_table_id
    '''

    def associateMetadataWithFiles(
        self, schemaGenerator: SchemaGenerator, metadataManifestPath: str, datasetId: str, manifest_record_type: str = 'both', 
        useSchemaLabel: bool = True, hideBlanks: bool = False, restrict_manifest = False,
    ) -> str:
        """Associate metadata with files in a storage dataset already on Synapse.
        Upload metadataManifest in the storage dataset folder on Synapse as well. Return synapseId of the uploaded manifest file.
        
        If this is a new manifest there could be no Synapse entities associated with the rows of this manifest
        this may be due to data type (e.g. clinical data) being tabular
        and not requiring files; to utilize uniform interfaces downstream
        (i.e. fileviews), a Synapse entity (a folder) is created for each row
        and an entity column is added to the manifest containing the resulting
        entity IDs; a table is also created at present as an additional interface
        for downstream query and interaction with the data.

        Args:
            metadataManifestPath: path to csv containing a validated metadata manifest.
            The manifest should include a column entityId containing synapse IDs of files/entities to be associated with metadata, if that is applicable to the dataset type.
            Some datasets, e.g. clinical data, do not contain file id's, but data is stored in a table: one row per item.
            In this case, the system creates a file on Synapse for each row in the table (e.g. patient, biospecimen) and associates the columnset data as metadata/annotations to his file.
            datasetId: synapse ID of folder containing the dataset
            useSchemaLabel: Default is True - use the schema label. If False, uses the display label from the schema. Attribute display names in the schema must not only include characters that are not accepted by Synapse. Annotation names may only contain: letters, numbers, '_' and '.'.
            manifest_record_type: valid values are 'entity', 'table' or 'both'. Specifies whether to create entity ids and folders for each row in a manifest, a Synapse table to house the entire manifest or do both.
            hideBlanks: Default is false. Boolean flag that does not upload annotation keys with blank values when true. Uploads Annotation keys with empty string values when false.
        Returns:
            manifest_synapse_file_id: SynID of manifest csv uploaded to synapse.

        Raises:
            ValueError: manifest_record_type is not 'entity', 'table' or 'both'
            FileNotFoundError: Manifest file does not exist at provided path.

        """

        # Check that record type provided matches expected input.
        manifest_record_types = ['entity', 'table', 'both']
        try:
            manifest_record_type in manifest_record_types
        except ValueError as err:
            raise ValueError(
                f"manifest_record_type provided: {manifest_record_type}, is not one of the accepted "
                f"types: {manifest_record_types}"
            ) from err

        # read new manifest csv
        try:
            load_args={
                "dtype":"string",
            }
            manifest = load_df(metadataManifestPath, preserve_raw_input = True, **load_args) #HOTFIX: set preserve_raw_input to true to allow mixed type cols in table uploads schematic#870
        except FileNotFoundError as err:
            raise FileNotFoundError(
                f"No manifest file was found at this path: {metadataManifestPath}"
            ) from err

        # Add uuid for table updates and fill.
        if not "Uuid" in manifest.columns:
            manifest["Uuid"] = ''

        for idx,row in manifest.iterrows():
            if not row["Uuid"]:
                gen_uuid = str(uuid.uuid4())
                row["Uuid"] = gen_uuid
                manifest.loc[idx, 'Uuid'] = gen_uuid

        # add entityId as a column if not already there or
        # fill any blanks with an empty string.
        if not "entityId" in manifest.columns:
            manifest["entityId"] = ""
        else:
            manifest["entityId"].fillna("", inplace=True)

        # get a schema explorer object to ensure schema attribute names used in manifest are translated to schema labels for synapse annotations
        se = SchemaExplorer()

        # Create table name here.
        if 'Component' in manifest.columns:
            component_name = manifest['Component'][0].lower()
            table_name = component_name + '_synapse_storage_manifest_table'
        else:
            component_name = ''
            table_name = 'synapse_storage_manifest_table'

        # If specified, upload manifest as a table and get the SynID and manifest
        if manifest_record_type == 'table' or manifest_record_type == 'both':
            manifest_synapse_table_id, manifest, table_manifest = self.upload_format_manifest_table(
                                                        se, manifest, datasetId, table_name, restrict = restrict_manifest, useSchemaLabel=useSchemaLabel)
            
        # Iterate over manifest rows, create Synapse entities and store corresponding entity IDs in manifest if needed
        # also set metadata for each synapse entity as Synapse annotations
        for idx, row in manifest.iterrows():
            if not row["entityId"] and (manifest_record_type == 'entity' or 
                manifest_record_type == 'both'):
                # no entity exists for this row
                # so create one
                rowEntity = Folder(str(uuid.uuid4()), parent=datasetId)
                rowEntity = self.syn.store(rowEntity)
                entityId = rowEntity["id"]
                row["entityId"] = entityId
                manifest.loc[idx, "entityId"] = entityId
            elif not row["entityId"] and manifest_record_type == 'table':
                # If not using entityIds, fill with manifest_table_id so 
                row["entityId"] = manifest_synapse_table_id
                entityId = ''
            else:
                # get the entity id corresponding to this row
                entityId = row["entityId"]

            # Adding annotations to connected files.
            if entityId:
                # Format annotations for Synapse
                annos = self.format_row_annotations(se, schemaGenerator, row, entityId, useSchemaLabel, hideBlanks)

                if annos:
                # Store annotations for an entity folder
                    self.syn.set_annotations(annos)

        # Load manifest to synapse as a CSV File
        manifest_synapse_file_id = self.uplodad_manifest_file(manifest, metadataManifestPath, datasetId, restrict_manifest, component_name = component_name)
        
        # Set annotations for the file manifest.
        manifest_annotations = self.format_manifest_annotations(manifest, manifest_synapse_file_id)
        self.syn.set_annotations(manifest_annotations)

        logger.info("Associated manifest file with dataset on Synapse.")
        
        if manifest_record_type == 'table' or manifest_record_type == 'both':
            # Update manifest Synapse table with new entity id column.
            self.make_synapse_table(
                table_to_load = table_manifest,
                dataset_id = datasetId,
                existingTableId = manifest_synapse_table_id,
                table_name = table_name,
                update_col = 'Uuid',
                specify_schema = False,
                restrict = restrict_manifest
                )
            
            # Set annotations for the table manifest
            manifest_annotations = self.format_manifest_annotations(manifest, manifest_synapse_table_id)
            self.syn.set_annotations(manifest_annotations)

        return manifest_synapse_file_id

    def getTableAnnotations(self, table_id:str):
        """Generate dictionary of annotations for the given Synapse file.
        Synapse returns all custom annotations as lists since they
        can contain multiple values. In all cases, the values will
        be converted into strings and concatenated with ", ".

        Args:
            fileId (str): Synapse ID for dataset file.

        Returns:
            dict: Annotations as comma-separated strings.
        """
        try:
            entity = self.syn.get(table_id, downloadFile=False)
            is_table = entity.concreteType.endswith(".TableEntity")
            annotations_raw = entity.annotations
        except SynapseHTTPError:
            # If an error occurs with retrieving entity, skip it
            # This could be caused by a temporary file view that
            # was deleted since its ID was retrieved
            is_file, is_table = False, False

        # Skip anything that isn't a file or folder
        if not (is_table):
            return None

        annotations = self.getEntityAnnotations(table_id, entity, annotations_raw)

        return annotations

    def getFileAnnotations(self, fileId: str) -> Dict[str, str]:
        """Generate dictionary of annotations for the given Synapse file.
        Synapse returns all custom annotations as lists since they
        can contain multiple values. In all cases, the values will
        be converted into strings and concatenated with ", ".

        Args:
            fileId (str): Synapse ID for dataset file.

        Returns:
            dict: Annotations as comma-separated strings.
        """

        # Get entity metadata, including annotations
        try:
            entity = self.syn.get(fileId, downloadFile=False)
            is_file = entity.concreteType.endswith(".FileEntity")
            is_folder = entity.concreteType.endswith(".Folder")
            annotations_raw = entity.annotations
        except SynapseHTTPError:
            # If an error occurs with retrieving entity, skip it
            # This could be caused by a temporary file view that
            # was deleted since its ID was retrieved
            is_file, is_folder = False, False

        # Skip anything that isn't a file or folder
        if not (is_file or is_folder):
            return None

        annotations = self.getEntityAnnotations(fileId, entity, annotations_raw)

        return annotations

    def getEntityAnnotations(self, fileId, entity, annotations_raw):
        # Extract annotations from their lists and stringify. For example:
        # {'YearofBirth': [1980], 'author': ['bruno', 'milen', 'sujay']}
        annotations = dict()
        for key, vals in annotations_raw.items():
            if isinstance(vals, list) and len(vals) == 1:
                annotations[key] = str(vals[0])
            else:
                annotations[key] = ", ".join(str(v) for v in vals)

        # Add the file entity ID and eTag, which weren't lists
        assert fileId == entity.id, (
            "For some reason, the Synapse ID in the response doesn't match"
            "the Synapse ID sent in the request (via synapseclient)."
        )
        annotations["entityId"] = fileId
        annotations["eTag"] = entity.etag

        return annotations

    def getDatasetAnnotations(
        self, datasetId: str, fill_na: bool = True, force_batch: bool = False
    ) -> pd.DataFrame:
        """Generate table for annotations across all files in given dataset.

        Args:
            datasetId (str): Synapse ID for dataset folder.
            fill_na (bool): Whether to replace missing values with
                blank strings.
            force_batch (bool): Whether to force the function to use
                the batch mode, which uses a file view to retrieve
                annotations for a given dataset. Default to False
                unless there are more than 50 files in the dataset.

        Returns:
            pd.DataFrame: Table of annotations.
        """
        # Get all files in given dataset
        dataset_files = self.getFilesInStorageDataset(datasetId)

        # if there are no dataset files, there are no annotations
        # return None
        if not dataset_files:
            return pd.DataFrame()

        dataset_files_map = dict(dataset_files)
        dataset_file_ids, _ = list(zip(*dataset_files))

        # Get annotations for each file from Step 1
        # Batch mode
        try_batch = len(dataset_files) >= 50 or force_batch
        if try_batch:
            try:
                logger.info("Trying batch mode for retrieving Synapse annotations")
                table = self.getDatasetAnnotationsBatch(datasetId, dataset_file_ids)
            except (SynapseAuthenticationError, SynapseHTTPError):
                logger.info(
                    f"Unable to create a temporary file view bound to {datasetId}. "
                    "Defaulting to slower iterative retrieval of annotations."
                )
                # Default to the slower non-batch method
                logger.info("Batch mode failed (probably due to permission error)")
                try_batch = False

        # Non-batch mode
        if not try_batch:
            logger.info("Using slower (non-batch) sequential mode")
            records = [self.getFileAnnotations(i) for i in dataset_file_ids]
            # Remove any annotations for non-file/folders (stored as None)
            records = filter(None, records)
            table = pd.DataFrame.from_records(records)

        # Add filenames for the files that "survived" annotation retrieval
        filenames = [dataset_files_map[i] for i in table["entityId"]]

        if 'Filename' not in table.columns:
            table.insert(0, "Filename", filenames)

        # Ensure that entityId and eTag are at the end
        entity_ids = table.pop("entityId")
        etags = table.pop("eTag")
        table.insert(len(table.columns), "entityId", entity_ids)
        table.insert(len(table.columns), "eTag", etags)

        # Missing values are filled in with empty strings for Google Sheets
        if fill_na:
            table.fillna("", inplace=True)

        # Force all values as strings
        return table.astype(str)

    def getDatasetProject(self, datasetId: str) -> str:
        """Get parent project for a given dataset ID.

        Args:
            datasetId (str): Synapse entity ID (folder or project).

        Raises:
            ValueError: Raised if Synapse ID cannot be retrieved
            by the user or if it doesn't appear in the file view.

        Returns:
            str: The Synapse ID for the parent project.
        """
        # Subset main file view
        dataset_index = self.storageFileviewTable["id"] == datasetId
        dataset_row = self.storageFileviewTable[dataset_index]

        # Return `projectId` for given row if only one found
        if len(dataset_row) == 1:
            dataset_project = dataset_row["projectId"].values[0]
            return dataset_project

        # Otherwise, check if already project itself
        try:
            syn_object = self.syn.get(datasetId)
            if syn_object.properties["concreteType"].endswith("Project"):
                return datasetId
        except SynapseHTTPError:
            raise PermissionError(
                f"The given dataset ({datasetId}) isn't accessible with this "
                "user. This might be caused by a typo in the dataset Synapse ID."
            )

        # If not, then assume dataset not in file view
        raise LookupError (
            f"The given dataset ({datasetId}) doesn't appear in the "
            f"configured file view ({self.storageFileview}). This might "
            "mean that the file view's scope needs to be updated."
        )     

    def getDatasetAnnotationsBatch(
        self, datasetId: str, dataset_file_ids: Sequence[str] = None
    ) -> pd.DataFrame:
        """Generate table for annotations across all files in given dataset.
        This function uses a temporary file view to generate a table
        instead of iteratively querying for individual entity annotations.
        This function is expected to run much faster than
        `self.getDatasetAnnotationsBatch` on large datasets.

        Args:
            datasetId (str): Synapse ID for dataset folder.
            dataset_file_ids (Sequence[str]): List of Synapse IDs
                for dataset files/folders used to subset the table.

        Returns:
            pd.DataFrame: Table of annotations.
        """
        # Create data frame from annotations file view
        with DatasetFileView(datasetId, self.syn) as fileview:
            table = fileview.query()

        if dataset_file_ids:
            table = table.loc[table.index.intersection(dataset_file_ids)]

        table = table.reset_index(drop=True)

        return table

    def _get_table_schema_by_cname(self, table_schema):

        # assume no duplicate column names in the table
        table_schema_by_cname = {}

        for col_record in table_schema:

            #TODO clean up dictionary for compactness (e.g. remove redundant 'name' key)
            table_schema_by_cname[col_record["name"]] = col_record

        return table_schema_by_cname

    def make_synapse_table(self, 
            table_to_load: pd.DataFrame, 
            dataset_id: str, table_name: str, 
            existingTableId: str = None,
            update_col: str = 'entityId', 
            column_type_dictionary: Dict = {}, 
            specify_schema: bool = True, 
            restrict: bool = False, 
            manipulation: str = 'update') -> str:
        '''
        Make a synapse table for record based data

        Args:
            table_to_load (pd.DataFrame): table to upload to synapse
            dataset_id (str): synID for dataset related to manifest to be uploaded as table
            existingTableId (str): Optional, synID of existing table to upload to
            table_name (str): Name of the table that will be displayed on synapse
            update_col (str): Optional, if updating a table by aligning on index, column to use as indices
            column_type_dictionary (Dict): dictionary of column types
            specify_schema (bool):  specify a schema for the table at upload according to types in column_type_dictionary
            restrict (bool): set to True if access restrictions need to be imposed on table when stored on synapse, False otherwise 
            manipulation (str): type of manipulation to do if a table exists already. Can be either "update" or "replace". 
                Defaults to "update" to preserve old behavior

        Returns:
            str: synId of table uploaded to synapse

        '''
        if existingTableId:
            existing_table, existing_results = self.get_synapse_table(existingTableId)

            manipulation = manipulation.lower()
            if manipulation not in ['update', 'replace']:
                raise NotImplementedError(
                    "Currently, only 'update' and 'replace' table operations are supported."
                )

            # create/update a table corresponding to this dataset in this dataset's parent project
            # update_col is the column in the table that has a unique code that will allow Synapse to
            # locate its position in the old and new table.
            if manipulation == 'update':
                table_to_load = update_df(existing_table, table_to_load, update_col)
                # store table with existing etag data and impose restrictions as appropriate
                self.syn.store(Table(existingTableId, table_to_load, etag = existing_results.etag), isRestricted = restrict)

            elif manipulation == 'replace':
                # remove rows
                self.syn.delete(existing_results)
                # wait for row deletion to finish on synapse before getting empty table
                sleep(10)
                
                # removes all current columns
                current_table = self.syn.get(existingTableId)
                current_columns = self.syn.getTableColumns(current_table)
                for col in current_columns:
                    current_table.removeColumn(col)

                if not table_name:
                    table_name = datasetName + 'table'
                
                # Process columns according to manifest entries
                table_schema_by_cname = self._get_table_schema_by_cname(column_type_dictionary) 
                datasetParentProject = self.getDatasetProject(dataset_id)
                if specify_schema:
                    if column_type_dictionary == {}:
                        logger.error("Did not provide a column_type_dictionary.")
                    #create list of columns:
                    cols = []
                    
                    for col in table_to_load.columns:
                        
                        if col in table_schema_by_cname:
                            col_type = table_schema_by_cname[col]['columnType']
                            max_size = table_schema_by_cname[col]['maximumSize'] if 'maximumSize' in table_schema_by_cname[col].keys() else 100
                            max_list_len = 250
                            if max_size and max_list_len:
                                cols.append(Column(name=col, columnType=col_type, 
                                    maximumSize=max_size, maximumListLength=max_list_len))
                            elif max_size:
                                cols.append(Column(name=col, columnType=col_type, 
                                    maximumSize=max_size))
                            else:
                                cols.append(Column(name=col, columnType=col_type))
                        else:
                            
                            #TODO add warning that the given col was not found and it's max size is set to 100
                            cols.append(Column(name=col, columnType='STRING', maximumSize=100))
                    
                    # adds new columns to schema
                    for col in cols:
                        current_table.addColumn(col)
                    self.syn.store(current_table, isRestricted = restrict)

                    # wait for synapse store to finish
                    sleep(1)

                    # build schema and table from columns and store with necessary restrictions
                    schema = Schema(name=table_name, columns=cols, parent=datasetParentProject)
                    schema.id = existingTableId
                    table = Table(schema, table_to_load, etag = existing_results.etag)
                    table = self.syn.store(table, isRestricted = restrict)
                else:
                    logging.error("Must specify a schema for table replacements")

            # remove system metadata from manifest
            existing_table.drop(columns = ['ROW_ID', 'ROW_VERSION'], inplace = True)
            return existingTableId
        else:
            datasetEntity = self.syn.get(dataset_id, downloadFile = False)
            datasetName = datasetEntity.name
            table_schema_by_cname = self._get_table_schema_by_cname(column_type_dictionary) 

            if not table_name:
                table_name = datasetName + 'table'
            datasetParentProject = self.getDatasetProject(dataset_id)
            if specify_schema:
                if column_type_dictionary == {}:
                    logger.error("Did not provide a column_type_dictionary.")
                #create list of columns:
                cols = []
                for col in table_to_load.columns:
                    if col in table_schema_by_cname:
                        col_type = table_schema_by_cname[col]['columnType']
                        max_size = table_schema_by_cname[col]['maximumSize'] if 'maximumSize' in table_schema_by_cname[col].keys() else 100
                        max_list_len = 250
                        if max_size and max_list_len:
                            cols.append(Column(name=col, columnType=col_type, 
                                maximumSize=max_size, maximumListLength=max_list_len))
                        elif max_size:
                            cols.append(Column(name=col, columnType=col_type, 
                                maximumSize=max_size))
                        else:
                            cols.append(Column(name=col, columnType=col_type))
                    else:
                        #TODO add warning that the given col was not found and it's max size is set to 100
                        cols.append(Column(name=col, columnType='STRING', maximumSize=100))
                schema = Schema(name=table_name, columns=cols, parent=datasetParentProject)
                table = Table(schema, table_to_load)
                table = self.syn.store(table, isRestricted = restrict)
                return table.schema.id
            else:
                # For just uploading the tables to synapse using default
                # column types.
                table = build_table(table_name, datasetParentProject, table_to_load)
                table = self.syn.store(table, isRestricted = restrict)
                return table.schema.id


class DatasetFileView:
    """Helper class to create temporary dataset file views.
    This class can be used in conjunction with a 'with' statement.
    This will ensure that the file view is deleted automatically.
    See SynapseStorage.getDatasetAnnotationsBatch for example usage.
    """

    def __init__(
        self,
        datasetId: str,
        synapse: Synapse,
        name: str = None,
        temporary: bool = True,
        parentId: str = None,
    ) -> None:
        """Create a file view scoped to a dataset folder.

        Args:
            datasetId (str): Synapse ID for a dataset folder/project.
            synapse (Synapse): Used for Synapse requests.
            name (str): Name of the file view (temporary or not).
            temporary (bool): Whether to delete the file view on exit
                of either a 'with' statement or Python entirely.
            parentId (str, optional): Synapse ID specifying where to
                store the file view. Defaults to datasetId.
        """

        self.datasetId = datasetId
        self.synapse = synapse
        self.is_temporary = temporary

        if name is None:
            self.name = f"schematic annotation file view for {self.datasetId}"

        if self.is_temporary:
            uid = secrets.token_urlsafe(5)
            self.name = f"{self.name} - UID {uid}"

        # TODO: Allow a DCC admin to configure a "universal parent"
        #       Such as a Synapse project writeable by everyone.
        self.parentId = datasetId if parentId is None else parentId

        # TODO: Create local sharing setting to hide from everyone else
        view_schema = EntityViewSchema(
            name=self.name,
            parent=self.parentId,
            scopes=self.datasetId,
            includeEntityTypes=[EntityViewType.FILE, EntityViewType.FOLDER],
            addDefaultViewColumns=False,
            addAnnotationColumns=True,
        )

        # TODO: Handle failure due to insufficient permissions by
        #       creating a temporary new project to store view
        self.view_schema = self.synapse.store(view_schema)

        # These are filled in after calling `self.query()`
        self.results = None
        self.table = None

        # Ensure deletion of the file view (last resort)
        if self.is_temporary:
            atexit.register(self.delete)

    def __enter__(self):
        """Return file view when entering 'with' statement."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Delete file view when exiting 'with' statement."""
        if self.is_temporary:
            self.delete()

    def delete(self):
        """Delete the file view on Synapse without deleting local table."""
        if self.view_schema is not None:
            self.synapse.delete(self.view_schema)
            self.view_schema = None

    def query(self, tidy=True, force=False):
        """Retrieve file view as a data frame (raw format sans index)."""
        if self.table is None or force:
            fileview_id = self.view_schema["id"]
            self.results = self.synapse.tableQuery(f"select * from {fileview_id}")
            self.table = self.results.asDataFrame(rowIdAndVersionInIndex=False)
        if tidy:
            self.tidy_table()
        return self.table

    def tidy_table(self):
        """Convert raw file view data frame into more usable format."""
        assert self.table is not None, "Must call `self.query()` first."
        self._fix_default_columns()
        self._fix_list_columns()
        self._fix_int_columns()
        return self.table

    def _fix_default_columns(self):
        """Rename default columns to match schematic expectations."""

        # Drop ROW_VERSION column if present
        if "ROW_VERSION" in self.table:
            del self.table["ROW_VERSION"]

        # Rename id column to entityId and set as data frame index
        if "ROW_ID" in self.table:
            self.table["entityId"] = "syn" + self.table["ROW_ID"].astype(str)
            self.table = self.table.set_index("entityId", drop=False)
            del self.table["ROW_ID"]

        # Rename ROW_ETAG column to eTag and place at end of data frame
        if "ROW_ETAG" in self.table:
            row_etags = self.table.pop("ROW_ETAG")
            self.table.insert(len(self.table.columns), "eTag", row_etags)

        return self.table

    def _get_columns_of_type(self, types):
        """Helper function to get list of columns of a given type(s)."""
        matching_columns = []
        for header in self.results.headers:
            if header.columnType in types:
                matching_columns.append(header.name)
        return matching_columns

    def _fix_list_columns(self):
        """Fix formatting of list-columns."""
        list_types = {"STRING_LIST", "INTEGER_LIST", "BOOLEAN_LIST"}
        list_columns = self._get_columns_of_type(list_types)
        for col in list_columns:
            self.table[col] = self.table[col].apply(lambda x: ", ".join(x))
        return self.table

    def _fix_int_columns(self):
        """Ensure that integer-columns are actually integers."""
        int_columns = self._get_columns_of_type({"INTEGER"})
        for col in int_columns:
            # Coercing to string because NaN is a floating point value
            # and cannot exist alongside integers in a column
            to_int_fn = lambda x: "" if np.isnan(x) else str(int(x))
            self.table[col] = self.table[col].apply(to_int_fn)
        return self.table
